"""
Tests for the assembler and disassembler.

These tests verify that:
1. Directives can be serialized and deserialized correctly (round-trip)
2. The assembler can parse bytecode text and produce correct directives
3. The disassembler can convert directives back to text
4. The full round-trip (text -> directives -> binary -> directives -> text) works
"""

import pytest
from dataclasses import fields

from fpy.bytecode.assembler import (
    parse as fpybc_parse,
    assemble,
    directives_to_fpybc,
)
from fpy.bytecode.directives import (
    Directive,
    # Directives with no args
    NoOpDirective,
    WaitRelDirective,
    WaitAbsDirective,
    OrDirective,
    AndDirective,
    IntEqualDirective,
    IntNotEqualDirective,
    UnsignedLessThanDirective,
    UnsignedLessThanOrEqualDirective,
    UnsignedGreaterThanDirective,
    UnsignedGreaterThanOrEqualDirective,
    SignedLessThanDirective,
    SignedLessThanOrEqualDirective,
    SignedGreaterThanDirective,
    SignedGreaterThanOrEqualDirective,
    FloatGreaterThanOrEqualDirective,
    FloatLessThanOrEqualDirective,
    FloatLessThanDirective,
    FloatGreaterThanDirective,
    FloatEqualDirective,
    FloatNotEqualDirective,
    NotDirective,
    FloatTruncateDirective,
    FloatExtendDirective,
    FloatToSignedIntDirective,
    SignedIntToFloatDirective,
    FloatToUnsignedIntDirective,
    UnsignedIntToFloatDirective,
    ExitDirective,
    PushTimeDirective,
    CallDirective,
    PeekDirective,
    IntAddDirective,
    IntSubtractDirective,
    IntMultiplyDirective,
    UnsignedIntDivideDirective,
    SignedIntDivideDirective,
    FloatAddDirective,
    FloatSubtractDirective,
    FloatMultiplyDirective,
    FloatDivideDirective,
    FloatExponentDirective,
    FloatLogDirective,
    FloatModuloDirective,
    SignedModuloDirective,
    UnsignedModuloDirective,
    IntegerSignedExtend8To64Directive,
    IntegerSignedExtend16To64Directive,
    IntegerSignedExtend32To64Directive,
    IntegerZeroExtend8To64Directive,
    IntegerZeroExtend16To64Directive,
    IntegerZeroExtend32To64Directive,
    IntegerTruncate64To8Directive,
    IntegerTruncate64To16Directive,
    IntegerTruncate64To32Directive,
    # Directives with args
    AllocateDirective,
    StoreRelDirective,
    StoreRelConstOffsetDirective,
    StoreAbsDirective,
    StoreAbsConstOffsetDirective,
    LoadRelDirective,
    LoadAbsDirective,
    DiscardDirective,
    PushValDirective,
    ConstCmdDirective,
    GotoDirective,
    IfDirective,
    PushTlmValDirective,
    PushPrmDirective,
    StackCmdDirective,
    MemCompareDirective,
    GetFieldDirective,
    ReturnDirective,
)
from fpy.types import serialize_directives, deserialize_directives


class TestDirectiveSerializationRoundTrip:
    """Test that each directive can be serialized and deserialized correctly."""

    def test_no_op(self):
        original = NoOpDirective()
        self._test_roundtrip(original)

    def test_wait_rel(self):
        original = WaitRelDirective()
        self._test_roundtrip(original)

    def test_wait_abs(self):
        original = WaitAbsDirective()
        self._test_roundtrip(original)

    def test_exit(self):
        original = ExitDirective()
        self._test_roundtrip(original)

    def test_push_time(self):
        original = PushTimeDirective()
        self._test_roundtrip(original)

    def test_call(self):
        original = CallDirective()
        self._test_roundtrip(original)

    def test_peek(self):
        original = PeekDirective()
        self._test_roundtrip(original)

    # Boolean operators
    def test_or(self):
        original = OrDirective()
        self._test_roundtrip(original)

    def test_and(self):
        original = AndDirective()
        self._test_roundtrip(original)

    def test_not(self):
        original = NotDirective()
        self._test_roundtrip(original)

    # Integer comparison operators
    def test_int_equal(self):
        original = IntEqualDirective()
        self._test_roundtrip(original)

    def test_int_not_equal(self):
        original = IntNotEqualDirective()
        self._test_roundtrip(original)

    def test_unsigned_less_than(self):
        original = UnsignedLessThanDirective()
        self._test_roundtrip(original)

    def test_unsigned_less_than_or_equal(self):
        original = UnsignedLessThanOrEqualDirective()
        self._test_roundtrip(original)

    def test_unsigned_greater_than(self):
        original = UnsignedGreaterThanDirective()
        self._test_roundtrip(original)

    def test_unsigned_greater_than_or_equal(self):
        original = UnsignedGreaterThanOrEqualDirective()
        self._test_roundtrip(original)

    def test_signed_less_than(self):
        original = SignedLessThanDirective()
        self._test_roundtrip(original)

    def test_signed_less_than_or_equal(self):
        original = SignedLessThanOrEqualDirective()
        self._test_roundtrip(original)

    def test_signed_greater_than(self):
        original = SignedGreaterThanDirective()
        self._test_roundtrip(original)

    def test_signed_greater_than_or_equal(self):
        original = SignedGreaterThanOrEqualDirective()
        self._test_roundtrip(original)

    # Float comparison operators
    def test_float_equal(self):
        original = FloatEqualDirective()
        self._test_roundtrip(original)

    def test_float_not_equal(self):
        original = FloatNotEqualDirective()
        self._test_roundtrip(original)

    def test_float_less_than(self):
        original = FloatLessThanDirective()
        self._test_roundtrip(original)

    def test_float_less_than_or_equal(self):
        original = FloatLessThanOrEqualDirective()
        self._test_roundtrip(original)

    def test_float_greater_than(self):
        original = FloatGreaterThanDirective()
        self._test_roundtrip(original)

    def test_float_greater_than_or_equal(self):
        original = FloatGreaterThanOrEqualDirective()
        self._test_roundtrip(original)

    # Integer arithmetic operators
    def test_int_add(self):
        original = IntAddDirective()
        self._test_roundtrip(original)

    def test_int_subtract(self):
        original = IntSubtractDirective()
        self._test_roundtrip(original)

    def test_int_multiply(self):
        original = IntMultiplyDirective()
        self._test_roundtrip(original)

    def test_unsigned_int_divide(self):
        original = UnsignedIntDivideDirective()
        self._test_roundtrip(original)

    def test_signed_int_divide(self):
        original = SignedIntDivideDirective()
        self._test_roundtrip(original)

    def test_signed_modulo(self):
        original = SignedModuloDirective()
        self._test_roundtrip(original)

    def test_unsigned_modulo(self):
        original = UnsignedModuloDirective()
        self._test_roundtrip(original)

    # Float arithmetic operators
    def test_float_add(self):
        original = FloatAddDirective()
        self._test_roundtrip(original)

    def test_float_subtract(self):
        original = FloatSubtractDirective()
        self._test_roundtrip(original)

    def test_float_multiply(self):
        original = FloatMultiplyDirective()
        self._test_roundtrip(original)

    def test_float_divide(self):
        original = FloatDivideDirective()
        self._test_roundtrip(original)

    def test_float_exponent(self):
        original = FloatExponentDirective()
        self._test_roundtrip(original)

    def test_float_log(self):
        original = FloatLogDirective()
        self._test_roundtrip(original)

    def test_float_modulo(self):
        original = FloatModuloDirective()
        self._test_roundtrip(original)

    # Type conversion operators
    def test_float_truncate(self):
        original = FloatTruncateDirective()
        self._test_roundtrip(original)

    def test_float_extend(self):
        original = FloatExtendDirective()
        self._test_roundtrip(original)

    def test_float_to_signed_int(self):
        original = FloatToSignedIntDirective()
        self._test_roundtrip(original)

    def test_signed_int_to_float(self):
        original = SignedIntToFloatDirective()
        self._test_roundtrip(original)

    def test_float_to_unsigned_int(self):
        original = FloatToUnsignedIntDirective()
        self._test_roundtrip(original)

    def test_unsigned_int_to_float(self):
        original = UnsignedIntToFloatDirective()
        self._test_roundtrip(original)

    # Integer extension/truncation
    def test_siext_8_64(self):
        original = IntegerSignedExtend8To64Directive()
        self._test_roundtrip(original)

    def test_siext_16_64(self):
        original = IntegerSignedExtend16To64Directive()
        self._test_roundtrip(original)

    def test_siext_32_64(self):
        original = IntegerSignedExtend32To64Directive()
        self._test_roundtrip(original)

    def test_ziext_8_64(self):
        original = IntegerZeroExtend8To64Directive()
        self._test_roundtrip(original)

    def test_ziext_16_64(self):
        original = IntegerZeroExtend16To64Directive()
        self._test_roundtrip(original)

    def test_ziext_32_64(self):
        original = IntegerZeroExtend32To64Directive()
        self._test_roundtrip(original)

    def test_itrunc_64_8(self):
        original = IntegerTruncate64To8Directive()
        self._test_roundtrip(original)

    def test_itrunc_64_16(self):
        original = IntegerTruncate64To16Directive()
        self._test_roundtrip(original)

    def test_itrunc_64_32(self):
        original = IntegerTruncate64To32Directive()
        self._test_roundtrip(original)

    # Directives with arguments
    def test_allocate(self):
        original = AllocateDirective(size=100)
        self._test_roundtrip(original)

    def test_allocate_zero(self):
        original = AllocateDirective(size=0)
        self._test_roundtrip(original)

    def test_allocate_large(self):
        original = AllocateDirective(size=0xFFFFFFFF)
        self._test_roundtrip(original)

    def test_store_rel(self):
        original = StoreRelDirective(size=8)
        self._test_roundtrip(original)

    def test_store_rel_const_offset(self):
        original = StoreRelConstOffsetDirective(lvar_offset=16, size=4)
        self._test_roundtrip(original)

    def test_store_rel_const_offset_negative(self):
        original = StoreRelConstOffsetDirective(lvar_offset=-8, size=4)
        self._test_roundtrip(original)

    def test_store_abs(self):
        original = StoreAbsDirective(size=8)
        self._test_roundtrip(original)

    def test_store_abs_const_offset(self):
        original = StoreAbsConstOffsetDirective(global_offset=100, size=4)
        self._test_roundtrip(original)

    def test_load_rel(self):
        original = LoadRelDirective(lvar_offset=0, size=8)
        self._test_roundtrip(original)

    def test_load_rel_negative_offset(self):
        original = LoadRelDirective(lvar_offset=-16, size=4)
        self._test_roundtrip(original)

    def test_load_abs(self):
        original = LoadAbsDirective(global_offset=50, size=8)
        self._test_roundtrip(original)

    def test_discard(self):
        original = DiscardDirective(size=4)
        self._test_roundtrip(original)

    def test_push_val_empty(self):
        original = PushValDirective(val=b"")
        self._test_roundtrip(original)

    def test_push_val_single_byte(self):
        original = PushValDirective(val=b"\x42")
        self._test_roundtrip(original)

    def test_push_val_multiple_bytes(self):
        original = PushValDirective(val=b"\x00\x01\x02\x03\x04\x05\x06\x07")
        self._test_roundtrip(original)

    def test_push_val_all_byte_values(self):
        original = PushValDirective(val=bytes(range(256)))
        self._test_roundtrip(original)

    def test_const_cmd(self):
        original = ConstCmdDirective(cmd_opcode=123, args=b"\x01\x02\x03")
        self._test_roundtrip(original)

    def test_const_cmd_empty_args(self):
        original = ConstCmdDirective(cmd_opcode=456, args=b"")
        self._test_roundtrip(original)

    def test_goto(self):
        original = GotoDirective(dir_idx=10)
        self._test_roundtrip(original)

    def test_goto_zero(self):
        original = GotoDirective(dir_idx=0)
        self._test_roundtrip(original)

    def test_if(self):
        original = IfDirective(false_goto_dir_index=5)
        self._test_roundtrip(original)

    def test_push_tlm_val(self):
        original = PushTlmValDirective(chan_id=100)
        self._test_roundtrip(original)

    def test_push_prm(self):
        original = PushPrmDirective(prm_id=200)
        self._test_roundtrip(original)

    def test_stack_cmd(self):
        original = StackCmdDirective(args_size=16)
        self._test_roundtrip(original)

    def test_memcmp(self):
        original = MemCompareDirective(size=32)
        self._test_roundtrip(original)

    def test_get_field(self):
        original = GetFieldDirective(parent_size=64, member_size=8)
        self._test_roundtrip(original)

    def test_return(self):
        original = ReturnDirective(return_val_size=8, call_args_size=16)
        self._test_roundtrip(original)

    def _test_roundtrip(self, original: Directive):
        """Helper to test serialize/deserialize round-trip for a single directive."""
        dirs = [original]
        serialized, _ = serialize_directives(dirs)
        deserialized = deserialize_directives(serialized)
        
        assert len(deserialized) == 1
        result = deserialized[0]
        
        # Check type matches
        assert type(result) == type(original)
        
        # Check all fields match
        for field in fields(original):
            if field.name in ('meta', 'id'):
                continue
            original_val = getattr(original, field.name)
            result_val = getattr(result, field.name)
            assert original_val == result_val, f"Field {field.name}: {original_val} != {result_val}"


class TestAssemblerParsing:
    """Test that the assembler can parse bytecode text correctly."""

    def test_parse_no_op(self):
        text = "no_op\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], NoOpDirective)

    def test_parse_multiple_no_args(self):
        text = """
no_op
exit
add
sub
"""
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 4
        assert isinstance(dirs[0], NoOpDirective)
        assert isinstance(dirs[1], ExitDirective)
        assert isinstance(dirs[2], IntAddDirective)
        assert isinstance(dirs[3], IntSubtractDirective)

    def test_parse_allocate(self):
        text = "allocate 100\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], AllocateDirective)
        assert dirs[0].size == 100

    def test_parse_load_rel(self):
        text = "load_rel -8 4\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], LoadRelDirective)
        assert dirs[0].lvar_offset == -8
        assert dirs[0].size == 4

    def test_parse_store_rel_const_offset(self):
        text = "store_rel_const_offset -16 8\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], StoreRelConstOffsetDirective)
        assert dirs[0].lvar_offset == -16
        assert dirs[0].size == 8

    def test_parse_push_val_with_bytes(self):
        text = "push_val 1 2 3 4\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], PushValDirective)
        assert dirs[0].val == b"\x01\x02\x03\x04"

    def test_parse_push_val_with_hex(self):
        text = "push_val 0xFF 0x00 0xAB\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], PushValDirective)
        assert dirs[0].val == b"\xff\x00\xab"

    def test_parse_push_val_empty(self):
        text = "push_val\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], PushValDirective)
        assert dirs[0].val == b""

    def test_parse_const_cmd(self):
        text = "const_cmd 123 1 2 3\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], ConstCmdDirective)
        assert dirs[0].cmd_opcode == 123
        assert dirs[0].args == b"\x01\x02\x03"

    def test_parse_goto_with_index(self):
        text = "goto 5\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], GotoDirective)
        assert dirs[0].dir_idx == 5

    def test_parse_goto_with_tag(self):
        text = """
goto end
no_op
end:
exit
"""
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 3
        assert isinstance(dirs[0], GotoDirective)
        assert dirs[0].dir_idx == 2  # Points to the 'exit' instruction

    def test_parse_if_with_tag(self):
        text = """
if skip
no_op
skip:
exit
"""
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 3
        assert isinstance(dirs[0], IfDirective)
        assert dirs[0].false_goto_dir_index == 2

    def test_parse_comments(self):
        text = """
# This is a comment
no_op  # inline comment
# another comment
exit
"""
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 2

    def test_parse_all_no_arg_ops(self):
        """Test that all no-arg operations can be parsed."""
        ops = [
            "no_op", "siext_8_64", "siext_16_64", "siext_32_64",
            "ziext_8_64", "ziext_16_64", "ziext_32_64",
            "itrunc_64_8", "itrunc_64_16", "itrunc_64_32",
            "fmod", "smod", "umod", "add", "sub", "mul", "udiv", "sdiv",
            "fadd", "fsub", "fmul", "fpow", "fdiv", "flog",
            "wait_rel", "wait_abs", "or", "and",
            "ieq", "ine", "ult", "ule", "ugt", "uge",
            "slt", "sle", "sgt", "sge",
            "fge", "fle", "flt", "fgt", "feq", "fne",
            "not", "fptrunc", "fpext", "fptosi", "sitofp", "fptoui", "uitofp",
            "exit", "push_time", "call", "peek",
        ]
        for op in ops:
            text = f"{op}\n"
            body = fpybc_parse(text)
            dirs = assemble(body)
            assert len(dirs) == 1, f"Failed for {op}"

    def test_parse_get_field(self):
        text = "get_field 64 8\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], GetFieldDirective)
        assert dirs[0].parent_size == 64
        assert dirs[0].member_size == 8

    def test_parse_return(self):
        text = "return 8 16\n"
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 1
        assert isinstance(dirs[0], ReturnDirective)
        assert dirs[0].return_val_size == 8
        assert dirs[0].call_args_size == 16


class TestDisassembler:
    """Test that directives_to_fpybc produces correct output."""

    def test_disassemble_no_op(self):
        dirs = [NoOpDirective()]
        text = directives_to_fpybc(dirs)
        assert text.strip() == "no_op"

    def test_disassemble_allocate(self):
        dirs = [AllocateDirective(size=100)]
        text = directives_to_fpybc(dirs)
        assert text.strip() == "allocate 100"

    def test_disassemble_load_rel(self):
        dirs = [LoadRelDirective(lvar_offset=-8, size=4)]
        text = directives_to_fpybc(dirs)
        assert text.strip() == "load_rel -8 4"

    def test_disassemble_push_val(self):
        dirs = [PushValDirective(val=b"\x01\x02\x03")]
        text = directives_to_fpybc(dirs)
        assert text.strip() == "push_val 1 2 3"

    def test_disassemble_goto(self):
        dirs = [GotoDirective(dir_idx=10)]
        text = directives_to_fpybc(dirs)
        assert text.strip() == "goto 10"

    def test_disassemble_const_cmd(self):
        dirs = [ConstCmdDirective(cmd_opcode=123, args=b"\x01\x02")]
        text = directives_to_fpybc(dirs)
        assert text.strip() == "const_cmd 123 1 2"

    def test_disassemble_multiple(self):
        dirs = [
            AllocateDirective(size=16),
            NoOpDirective(),
            ExitDirective(),
        ]
        text = directives_to_fpybc(dirs)
        lines = [line for line in text.strip().split('\n') if line]
        assert len(lines) == 3
        assert lines[0] == "allocate 16"
        assert lines[1] == "no_op"
        assert lines[2] == "exit"


class TestAssemblerDisassemblerRoundTrip:
    """Test that text -> directives -> text produces equivalent results."""

    def test_roundtrip_no_op(self):
        self._test_text_roundtrip("no_op\n")

    def test_roundtrip_allocate(self):
        self._test_text_roundtrip("allocate 100\n")

    def test_roundtrip_load_rel(self):
        self._test_text_roundtrip("load_rel -8 4\n")

    def test_roundtrip_push_val(self):
        self._test_text_roundtrip("push_val 1 2 3\n")

    def test_roundtrip_goto(self):
        self._test_text_roundtrip("goto 5\n")

    def test_roundtrip_if(self):
        self._test_text_roundtrip("if 10\n")

    def test_roundtrip_complex_sequence(self):
        text = """allocate 16
load_rel -8 4
push_val 1 2 3 4
add
store_rel 4
goto 0
exit
"""
        self._test_text_roundtrip(text)

    def test_roundtrip_all_no_arg_ops(self):
        """Test round-trip for all no-arg operations."""
        ops = [
            "no_op", "siext_8_64", "siext_16_64", "siext_32_64",
            "ziext_8_64", "ziext_16_64", "ziext_32_64",
            "itrunc_64_8", "itrunc_64_16", "itrunc_64_32",
            "fmod", "smod", "umod", "add", "sub", "mul", "udiv", "sdiv",
            "fadd", "fsub", "fmul", "fpow", "fdiv", "flog",
            "wait_rel", "wait_abs", "or", "and",
            "ieq", "ine", "ult", "ule", "ugt", "uge",
            "slt", "sle", "sgt", "sge",
            "fge", "fle", "flt", "fgt", "feq", "fne",
            "not", "fptrunc", "fpext", "fptosi", "sitofp", "fptoui", "uitofp",
            "exit", "push_time", "call", "peek",
        ]
        for op in ops:
            self._test_text_roundtrip(f"{op}\n")

    def _test_text_roundtrip(self, original_text: str):
        """Helper to test text -> directives -> text round-trip."""
        # Parse and assemble
        body = fpybc_parse(original_text)
        dirs = assemble(body)
        
        # Disassemble back to text
        result_text = directives_to_fpybc(dirs)
        
        # Parse again and compare directives
        body2 = fpybc_parse(result_text)
        dirs2 = assemble(body2)
        
        assert len(dirs) == len(dirs2)
        for d1, d2 in zip(dirs, dirs2):
            assert type(d1) == type(d2)
            for field in fields(d1):
                if field.name in ('meta', 'id'):
                    continue
                v1 = getattr(d1, field.name)
                v2 = getattr(d2, field.name)
                assert v1 == v2, f"{field.name}: {v1} != {v2}"


class TestFullRoundTrip:
    """Test the complete round-trip: text -> directives -> binary -> directives -> text."""

    def test_full_roundtrip_simple(self):
        text = "no_op\nexit\n"
        self._test_full_roundtrip(text)

    def test_full_roundtrip_with_args(self):
        text = """allocate 32
load_rel -16 8
push_val 0 1 2 3 4 5 6 7
add
store_rel_const_offset -8 8
discard 8
exit
"""
        self._test_full_roundtrip(text)

    def test_full_roundtrip_control_flow(self):
        text = """allocate 8
goto 3
no_op
if 5
no_op
exit
"""
        self._test_full_roundtrip(text)

    def test_full_roundtrip_arithmetic(self):
        text = """add
sub
mul
udiv
sdiv
fadd
fsub
fmul
fdiv
fpow
flog
fmod
smod
umod
"""
        self._test_full_roundtrip(text)

    def test_full_roundtrip_comparisons(self):
        text = """ieq
ine
ult
ule
ugt
uge
slt
sle
sgt
sge
feq
fne
flt
fle
fgt
fge
"""
        self._test_full_roundtrip(text)

    def test_full_roundtrip_type_conversions(self):
        text = """siext_8_64
siext_16_64
siext_32_64
ziext_8_64
ziext_16_64
ziext_32_64
itrunc_64_8
itrunc_64_16
itrunc_64_32
fptrunc
fpext
fptosi
sitofp
fptoui
uitofp
"""
        self._test_full_roundtrip(text)

    def test_full_roundtrip_const_cmd(self):
        text = "const_cmd 12345 0 1 2 3 4 5 6 7 8 9\n"
        self._test_full_roundtrip(text)

    def test_full_roundtrip_empty_push_val(self):
        text = "push_val\n"
        self._test_full_roundtrip(text)

    def test_full_roundtrip_large_push_val(self):
        # Generate a push_val with many bytes
        bytes_str = " ".join(str(i % 256) for i in range(100))
        text = f"push_val {bytes_str}\n"
        self._test_full_roundtrip(text)

    def _test_full_roundtrip(self, original_text: str):
        """Helper to test full round-trip: text -> dirs -> binary -> dirs -> text."""
        # Step 1: Parse and assemble original text to directives
        body = fpybc_parse(original_text)
        dirs = assemble(body)
        
        # Step 2: Serialize directives to binary
        binary, _ = serialize_directives(dirs)
        
        # Step 3: Deserialize binary back to directives
        dirs2 = deserialize_directives(binary)
        
        # Step 4: Convert directives back to text
        result_text = directives_to_fpybc(dirs2)
        
        # Step 5: Parse the result text and compare directives
        body3 = fpybc_parse(result_text)
        dirs3 = assemble(body3)
        
        # Compare original directives with final directives
        assert len(dirs) == len(dirs3), f"Length mismatch: {len(dirs)} != {len(dirs3)}"
        for i, (d1, d3) in enumerate(zip(dirs, dirs3)):
            assert type(d1) == type(d3), f"Type mismatch at {i}: {type(d1)} != {type(d3)}"
            for field in fields(d1):
                if field.name in ('meta', 'id'):
                    continue
                v1 = getattr(d1, field.name)
                v3 = getattr(d3, field.name)
                assert v1 == v3, f"Field {field.name} mismatch at directive {i}: {v1} != {v3}"


class TestEdgeCases:
    """Test edge cases and potential error conditions."""

    def test_unknown_goto_tag_fails(self):
        text = "goto unknown_tag\n"
        body = fpybc_parse(text)
        with pytest.raises(RuntimeError, match="Unknown tag"):
            assemble(body)

    def test_large_directive_index(self):
        # Test with a very large goto index
        dirs = [GotoDirective(dir_idx=0xFFFFFFFE)]
        serialized, _ = serialize_directives(dirs)
        deserialized = deserialize_directives(serialized)
        assert deserialized[0].dir_idx == 0xFFFFFFFE

    def test_negative_offset(self):
        dirs = [LoadRelDirective(lvar_offset=-2147483648, size=4)]  # min I32
        serialized, _ = serialize_directives(dirs)
        deserialized = deserialize_directives(serialized)
        assert deserialized[0].lvar_offset == -2147483648

    def test_max_positive_offset(self):
        dirs = [LoadRelDirective(lvar_offset=2147483647, size=4)]  # max I32
        serialized, _ = serialize_directives(dirs)
        deserialized = deserialize_directives(serialized)
        assert deserialized[0].lvar_offset == 2147483647

    def test_multiple_tags_same_location(self):
        text = """
tag1:
tag2:
no_op
goto tag1
goto tag2
"""
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 3
        assert dirs[1].dir_idx == 0
        assert dirs[2].dir_idx == 0

    def test_tag_at_end(self):
        text = """
no_op
goto end
exit
end:
"""
        body = fpybc_parse(text)
        dirs = assemble(body)
        assert len(dirs) == 3
        assert dirs[1].dir_idx == 3  # Points past the last instruction


class TestMultipleDirectives:
    """Test sequences with multiple directives."""

    def test_serialize_multiple_directives(self):
        dirs = [
            AllocateDirective(size=32),
            PushValDirective(val=b"\x01\x02\x03\x04"),
            StoreRelConstOffsetDirective(lvar_offset=-8, size=4),
            LoadRelDirective(lvar_offset=-8, size=4),
            IntAddDirective(),
            GotoDirective(dir_idx=3),
            ExitDirective(),
        ]
        serialized, _ = serialize_directives(dirs)
        deserialized = deserialize_directives(serialized)
        
        assert len(deserialized) == len(dirs)
        for orig, d in zip(dirs, deserialized):
            assert type(orig) == type(d)

    def test_empty_directive_list(self):
        dirs = []
        serialized, _ = serialize_directives(dirs)
        deserialized = deserialize_directives(serialized)
        assert len(deserialized) == 0

    def test_many_directives(self):
        # Create a sequence with many directives
        dirs = [NoOpDirective() for _ in range(100)]
        serialized, _ = serialize_directives(dirs)
        deserialized = deserialize_directives(serialized)
        assert len(deserialized) == 100
        assert all(isinstance(d, NoOpDirective) for d in deserialized)




from __future__ import annotations
from fpy.bytecode.directives import (
    ExitDirective,
    FloatLogDirective,
    PushTimeDirective,
    SignedIntToFloatDirective,
    WaitAbsDirective,
    WaitRelDirective,
)
from fpy.ir import Ir, IrIf, IrLabel
from fpy.syntax import Ast
from fpy.types import BuiltinSymbol, FpyStringValue, NothingValue
from fprime_gds.common.models.serialize.time_type import TimeType as TimeValue
from fprime_gds.common.models.serialize.numerical_types import (
    U8Type as U8Value,
    U16Type as U16Value,
    U32Type as U32Value,
    I64Type as I64Value,
    F64Type as F64Value,
)
from fpy.bytecode.directives import (
    FloatLessThanDirective,
    FloatMultiplyDirective,
    FloatSubtractDirective,
    FloatToUnsignedIntDirective,
    IntMultiplyDirective,
    IntegerTruncate64To32Directive,
    IntegerZeroExtend32To64Directive,
    PeekDirective,
    PushTimeDirective,
    PushValDirective,
    FloatLogDirective,
    Directive,
    ExitDirective,
    SignedLessThanDirective,
    StackSizeType,
    UnsignedIntToFloatDirective,
    WaitAbsDirective,
    WaitRelDirective,
)


def generate_abs_float(node: Ast) -> list[Directive | Ir]:
    # if input is < 0 multiply by -1
    leave_unmodified = IrLabel(node, "else")
    dirs = [
        # copy the f64
        PushValDirective(StackSizeType(8).serialize()),
        PushValDirective(StackSizeType(0).serialize()),
        PeekDirective(),
        # push 0
        PushValDirective(F64Value(0.0).serialize()),
        # check <
        FloatLessThanDirective(),
        IrIf(leave_unmodified),
        # push -1
        PushValDirective(F64Value(-1.0).serialize()),
        # and multiply
        FloatMultiplyDirective(),
        # otherwise do nothing
        leave_unmodified,
    ]
    return dirs


MACRO_ABS_FLOAT = BuiltinSymbol("abs", F64Value, [("value", F64Value, None)], generate_abs_float)


def generate_abs_signed_int(node: Ast) -> list[Directive | Ir]:
    # if input is < 0 multiply by -1
    leave_unmodified = IrLabel(node, "else")
    dirs = [
        # copy the I64
        PushValDirective(StackSizeType(8).serialize()),
        PushValDirective(StackSizeType(0).serialize()),
        PeekDirective(),
        # push 0
        PushValDirective(I64Value(0).serialize()),
        # check <
        SignedLessThanDirective(),
        IrIf(leave_unmodified),
        # push -1
        PushValDirective(I64Value(-1).serialize()),
        # and multiply
        IntMultiplyDirective(),
        # otherwise do nothing
        leave_unmodified,
    ]
    return dirs


MACRO_ABS_SIGNED_INT = BuiltinSymbol(
    "abs", I64Value, [("value", I64Value, None)], generate_abs_signed_int
)

MACRO_SLEEP_SECONDS_USECONDS = BuiltinSymbol(
    "sleep",
    NothingValue,
    [
        (
            "seconds",
            U32Value,
            U32Value(0),
        ),
        ("useconds", U32Value, U32Value(0)),
    ],
    lambda n: [WaitRelDirective()],
)


def generate_sleep_float(node: Ast) -> list[Directive | Ir]:
    # convert F64 to seconds and microseconds
    dirs = [
        # first do seconds
        # copy the f64
        PushValDirective(StackSizeType(8).serialize()),
        PushValDirective(StackSizeType(0).serialize()),
        PeekDirective(),
        # convert to U64
        FloatToUnsignedIntDirective(),
        # and then U32
        IntegerTruncate64To32Directive(),
        # now we have f64, u32 (seconds) on stack
        # now do microseconds
        # copy the f64 and u32
        PushValDirective(StackSizeType(12).serialize()),
        PushValDirective(StackSizeType(0).serialize()),
        PeekDirective(),
        # turn the u32 into a float
        IntegerZeroExtend32To64Directive(),
        UnsignedIntToFloatDirective(),
        # subtract, this should give us the frac
        FloatSubtractDirective(),
        # okay now multiply by 1000000
        PushValDirective(F64Value(1_000_000.0).serialize()),
        # now convert to u32
        FloatToUnsignedIntDirective(),
        IntegerTruncate64To32Directive(),
    ]

    return dirs


MACRO_SLEEP_FLOAT = BuiltinSymbol(
    "sleep", NothingValue, [("seconds", F64Value, None)], generate_sleep_float
)


def generate_log_signed_int(node: Ast) -> list[Directive | Ir]:
    return [
        # convert int to float
        SignedIntToFloatDirective(),
        FloatLogDirective(),
    ]


MACROS: dict[str, BuiltinSymbol] = {
    "sleep": MACRO_SLEEP_SECONDS_USECONDS,
    "sleep_until": BuiltinSymbol(
        "sleep_until",
        NothingValue,
        [("wakeup_time", TimeValue, None)],
        lambda n: [WaitAbsDirective()],
    ),
    "exit": BuiltinSymbol(
        "exit", NothingValue, [("exit_code", U8Value, None)], lambda n: [ExitDirective()]
    ),
    "log": BuiltinSymbol(
        "log", F64Value, [("operand", F64Value, None)], lambda n: [FloatLogDirective()]
    ),
    "now": BuiltinSymbol("now", TimeValue, [], lambda n: [PushTimeDirective()]),
    "iabs": MACRO_ABS_SIGNED_INT,
    "fabs": MACRO_ABS_FLOAT,
    # time() parses ISO 8601 timestamps at compile time
    # The generate function should never be called since this is always const-evaluated
    "time": BuiltinSymbol(
        "time",
        TimeValue,
        [
            ("timestamp", FpyStringValue, None),
            ("time_base", U16Value, U16Value(0)),
            ("time_context", U8Value, U8Value(0)),
        ],
        lambda n: [],  # placeholder - const eval handles this
    ),
}



from __future__ import annotations
from dataclasses import dataclass
from typing import Union
from fpy.syntax import Ast


class Ir:
    pass


class IrLabel(Ir):
    def __init__(self, node: Ast, label: str):
        super().__init__()
        self.name = f"{node.id}.{label}"

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, value):
        return isinstance(value, IrLabel) and value.name == self.name

    def __repr__(self):
        return f"IrLabel({self.name})"


@dataclass(frozen=True, unsafe_hash=True)
class IrGoto(Ir):
    label: IrLabel


@dataclass(frozen=True, unsafe_hash=True)
class IrIf(Ir):
    goto_if_false_label: IrLabel


@dataclass(frozen=True, unsafe_hash=True)
class IrPushLabelOffset(Ir):
    """pushes the label's offset from the start of the file to the stack, as a StackSizeType"""
    label: IrLabel


from types import SimpleNamespace

import pytest

from fpy.syntax import (
    AstAssign,
    AstBoolean,
    AstLiteral,
    AstNumber,
    AstPass,
    AstBlock,
    AstString,
    AstVar,
)
from fpy.types import Transformer, Visitor, TopDownVisitor


class _State(SimpleNamespace):
    def __init__(self):
        super().__init__(errors=[])

    def err(self, msg, node):  # pragma: no cover - forwarded through tests
        self.errors.append((msg, node))


def _make_state():
    return _State()


def test_visitor_registers_union_types():
    class LiteralVisitor(Visitor):
        def __init__(self):
            self.seen = []
            super().__init__()

        def visit_literals(self, node: AstLiteral, state):
            self.seen.append(type(node))

    visitor = LiteralVisitor()
    state = _make_state()

    string_node = AstString(meta=None, value="value")
    boolean_node = AstBoolean(meta=None, value=True)

    visitor.run(string_node, state)
    visitor.run(boolean_node, state)

    assert visitor.visitors.get(AstString).__func__ is LiteralVisitor.visit_literals
    assert visitor.visitors.get(AstBoolean).__func__ is LiteralVisitor.visit_literals
    assert visitor.seen == [AstString, AstBoolean]


def test_visitor_depth_first_traversal():
    class RecordingVisitor(Visitor):
        def __init__(self):
            self.visited = []
            super().__init__()

        def visit_var(self, node: AstVar, state):
            self.visited.append(f"var:{node.var}")

        def visit_number(self, node: AstNumber, state):
            self.visited.append(f"number:{node.value}")

        def visit_assign(self, node: AstAssign, state):
            self.visited.append("assign")

    visitor = RecordingVisitor()
    state = _make_state()

    lhs = AstVar(meta=None, var="x")
    rhs = AstNumber(meta=None, value=7)
    assign = AstAssign(meta=None, lhs=lhs, type_ann=None, rhs=rhs)

    visitor.run(assign, state)

    assert visitor.visited == ["var:x", "number:7", "assign"]

def test_top_down_visitor_breadth_first_order():
    class RecordingTopDownVisitor(TopDownVisitor):
        def __init__(self):
            self.visited = []
            super().__init__()

        def visit_assign(self, node: AstAssign, state):
            self.visited.append("assign")

        def visit_var(self, node: AstVar, state):
            self.visited.append(f"var:{node.var}")

        def visit_number(self, node: AstNumber, state):
            self.visited.append(f"number:{node.value}")

    visitor = RecordingTopDownVisitor()
    state = _make_state()

    lhs = AstVar(meta=None, var="x")
    rhs = AstNumber(meta=None, value=5)
    assign = AstAssign(meta=None, lhs=lhs, type_ann=None, rhs=rhs)

    visitor.run(assign, state)

    assert visitor.visited == ["assign", "var:x", "number:5"]


def test_visitor_default_handler_invoked():
    class DefaultingVisitor(Visitor):
        def __init__(self):
            self.default_hits = []
            super().__init__()

        def visit_default(self, node, state):
            self.default_hits.append(type(node))

        def visit_assign(self, node: AstAssign, state):
            pass

    visitor = DefaultingVisitor()
    state = _make_state()

    node = AstVar(meta=None, var="y")
    visitor.run(node, state)

    assert visitor.default_hits == [AstVar]


def test_visitor_stops_on_error():
    class ErroringVisitor(Visitor):
        def __init__(self):
            super().__init__()

        def visit_var(self, node: AstVar, state):
            state.err("boom", node)

        def visit_number(self, node: AstNumber, state):
            raise AssertionError("should not be reached")

    visitor = ErroringVisitor()
    state = _make_state()

    lhs = AstVar(meta=None, var="x")
    rhs = AstNumber(meta=None, value=9)
    assign = AstAssign(meta=None, lhs=lhs, type_ann=None, rhs=rhs)

    visitor.run(assign, state)

    assert state.errors == [("boom", lhs)]


def test_visitor_handler_exception_propagates():
    class ExplodingVisitor(Visitor):
        def visit_var(self, node: AstVar, state):
            raise RuntimeError("should fail")

    visitor = ExplodingVisitor()
    state = _make_state()

    node = AstVar(meta=None, var="z")

    with pytest.raises(RuntimeError, match="should fail"):
        visitor.run(node, state)

def test_transformer_replaces_child_nodes():
    class ZeroingTransformer(Transformer):
        def visit_number(self, node: AstNumber, state):
            if node.value == 42:
                return AstNumber(meta=node.meta, value=0)
            return None

    transformer = ZeroingTransformer()
    state = _make_state()

    lhs = AstVar(meta=None, var="answer")
    rhs = AstNumber(meta=None, value=42)
    assign = AstAssign(meta=None, lhs=lhs, type_ann=None, rhs=rhs)

    transformer.run(assign, state)

    assert isinstance(assign.rhs, AstNumber)
    assert assign.rhs is not rhs
    assert assign.rhs.value == 0


def test_transformer_deletes_nodes_from_lists():
    class DropPassesTransformer(Transformer):
        def visit_pass(self, node: AstPass, state):
            return Transformer.Delete

    transformer = DropPassesTransformer()
    state = _make_state()

    body = AstBlock(meta=None, stmts=[AstPass(meta=None), AstPass(meta=None)])

    transformer.run(body, state)

    assert body.stmts == []


def test_transformer_expands_nodes_in_lists():
    class ExpandPassTransformer(Transformer):
        def visit_pass(self, node: AstPass, state):
            first = AstPass(meta=node.meta)
            first.tag = "first"  # type: ignore[attr-defined]
            second = AstPass(meta=node.meta)
            second.tag = "second"  # type: ignore[attr-defined]
            return [first, second]

    transformer = ExpandPassTransformer()
    state = _make_state()

    original = AstPass(meta=None)
    body = AstBlock(meta=None, stmts=[original])

    transformer.run(body, state)

    assert len(body.stmts) == 2
    assert getattr(body.stmts[0], "tag") == "first"
    assert getattr(body.stmts[1], "tag") == "second"
    assert body.stmts[0] is not original
    assert body.stmts[1] is not original



from __future__ import annotations
from lark.tree import Meta
from fpy.bytecode.directives import BinaryStackOp, Directive, LoopVarType
from fpy.syntax import (
    Ast,
    AstAssign,
    AstBinaryOp,
    AstBoolean,
    AstFor,
    AstFuncCall,
    AstNumber,
    AstTypeExpr,
    AstRange,
    AstVar,
    AstWhile,
)
from fpy.types import (
    CompileState,
    ForLoopAnalysis,
    FppType,
    Symbol,
    FpyIntegerValue,
    Transformer,
)
from fprime_gds.common.models.serialize.type_base import BaseType as FppValue
from fprime_gds.common.models.serialize.bool_type import BoolType as BoolValue
from fprime_gds.common.models.serialize.numerical_types import IntegerType as IntegerValue


class DesugarForLoops(Transformer):

    # this function forces you to give values for all of these dicts
    # really the point is to make sure i don't forget to consider this
    # for one of the new nodes
    def new(
        self,
        state: CompileState,
        node: Ast,
        contextual_type: FppType | None,
        synthesized_type: FppType | None,
        contextual_value: FppValue | None,
        op_intermediate_type: type[Directive] | None,
        resolved_symbol: Symbol | None,
    ) -> Ast:
        node.id = state.next_node_id
        state.next_node_id += 1
        state.contextual_types[node] = contextual_type
        state.synthesized_types[node] = synthesized_type
        state.contextual_values[node] = contextual_value
        state.op_intermediate_types[node] = op_intermediate_type
        state.resolved_symbols[node] = resolved_symbol
        return node

    def initialize_loop_var(
        self, state: CompileState, loop_node: AstFor, loop_info: ForLoopAnalysis
    ) -> Ast:
        # 1 <node.loop_var>: LoopVarType = <node.range.lower_bound>
        # OR (depending on whether redeclaring or not)
        # 1 <node.loop_var> = <node.range.lower_bound>

        if loop_info.reuse_existing_loop_var:
            loop_var_type_var = None
        else:
            loop_var_type_name = LoopVarType.get_canonical_name()
            # create a new node for the type_ann
            loop_var_type_var = self.new(
                state,
                AstTypeExpr(None, [loop_var_type_name]),
                contextual_type=None,
                synthesized_type=None,
                contextual_value=None,
                op_intermediate_type=None,
                resolved_symbol=LoopVarType,
            )

        lhs = loop_node.loop_var
        rhs = loop_node.range.lower_bound
        return self.new(
            state,
            AstAssign(None, lhs, loop_var_type_var, rhs),
            contextual_type=None,
            synthesized_type=None,
            contextual_value=None,
            op_intermediate_type=None,
            resolved_symbol=None,
        )

    def declare_upper_bound_var(
        self, state: CompileState, loop_node: AstFor, loop_info: ForLoopAnalysis
    ) -> Ast:
        # 2 $upper_bound_var: LoopVarType = <node.range.upper_bound>

        # ub var for use in assignment
        # type is gonna be loop var type
        # gonna be used in the astassign lhs, no need assign a type in the dict
        upper_bound_var: AstVar = self.new(
            state,
            AstVar(None, loop_info.upper_bound_var.name),
            contextual_type=None,
            synthesized_type=None,
            contextual_value=None,
            op_intermediate_type=None,
            resolved_symbol=loop_info.upper_bound_var,
        )

        loop_var_type_name = LoopVarType.get_canonical_name()
        # create a new node for the type_ann
        loop_var_type_var = self.new(
            state,
            AstTypeExpr(None, [loop_var_type_name]),
            contextual_type=None,
            synthesized_type=None,
            contextual_value=None,
            op_intermediate_type=None,
            resolved_symbol=LoopVarType,
        )

        # assign ub to ub var
        # not an expr, not a symbol
        return self.new(
            state,
            AstAssign(
                None, upper_bound_var, loop_var_type_var, loop_node.range.upper_bound
            ),
            contextual_type=None,
            synthesized_type=None,
            contextual_value=None,
            op_intermediate_type=None,
            resolved_symbol=None,
        )

    def loop_var_plus_one(
        self, state: CompileState, loop_node: AstFor, loop_info: ForLoopAnalysis
    ):
        # <node.loop_var> + 1
        # the expression adding one to the lv
        # will have conv type of lv, unconverted type depends what the addition intermediate type is
        # we've already determined the dir
        lhs = self.new(
            state,
            AstVar(None, loop_info.loop_var.name),
            contextual_type=LoopVarType,
            synthesized_type=LoopVarType,
            contextual_value=None,
            op_intermediate_type=None,
            resolved_symbol=loop_info.loop_var,
        )
        rhs = self.new(
            state,
            AstNumber(None, 1),
            contextual_type=LoopVarType,
            synthesized_type=FpyIntegerValue,
            contextual_value=LoopVarType(1),
            op_intermediate_type=None,
            resolved_symbol=None,
        )

        return self.new(
            state,
            AstBinaryOp(None, lhs, BinaryStackOp.ADD, rhs),
            contextual_type=LoopVarType,
            synthesized_type=LoopVarType,
            contextual_value=None,
            op_intermediate_type=LoopVarType,
            resolved_symbol=None,
        )

    def increment_loop_var(
        self, state: CompileState, loop_node: AstFor, loop_info: ForLoopAnalysis
    ) -> Ast:
        # <node.loop_var> = <node.loop_var> + 1

        # create a new loop var symbol for use in lhs of loop var inc
        lhs = self.new(
            state,
            AstVar(None, loop_info.loop_var.name),
            contextual_type=None,
            synthesized_type=None,
            contextual_value=None,
            op_intermediate_type=None,
            resolved_symbol=loop_info.loop_var,
        )

        rhs = self.loop_var_plus_one(state, loop_node, loop_info)

        return self.new(
            state,
            AstAssign(None, lhs, None, rhs),
            contextual_type=None,
            synthesized_type=None,
            contextual_value=None,
            op_intermediate_type=None,
            resolved_symbol=None,
        )

    def while_loop_condition(
        self, state: CompileState, loop_node: AstFor, loop_info: ForLoopAnalysis
    ) -> Ast:
        # <node.loop_var> < $upper_bound_var
        # create a new loop var symbol for use in lhs
        lhs = self.new(
            state,
            AstVar(None, loop_info.loop_var.name),
            contextual_type=LoopVarType,
            synthesized_type=LoopVarType,
            contextual_value=None,
            op_intermediate_type=None,
            resolved_symbol=loop_info.loop_var,
        )
        rhs = self.new(
            state,
            AstVar(None, loop_info.upper_bound_var.name),
            contextual_type=LoopVarType,
            synthesized_type=LoopVarType,
            contextual_value=None,
            op_intermediate_type=None,
            resolved_symbol=loop_info.upper_bound_var,
        )

        return self.new(
            state,
            AstBinaryOp(None, lhs, BinaryStackOp.LESS_THAN, rhs),
            contextual_type=BoolValue,
            synthesized_type=BoolValue,
            contextual_value=None,
            op_intermediate_type=LoopVarType,
            resolved_symbol=None,
        )

    def while_loop(
        self, state: CompileState, loop_node: AstFor, loop_info: ForLoopAnalysis
    ) -> Ast:
        #  while <node.loop_var> < $upper_bound_var:
        #     <node.body>
        #     <node.loop_var> = <node.loop_var> + 1

        condition = self.while_loop_condition(state, loop_node, loop_info)
        increment = self.increment_loop_var(state, loop_node, loop_info)

        body = loop_node.body

        body.stmts.append(increment)

        return self.new(
            state,
            AstWhile(None, condition, body),
            contextual_type=None,
            synthesized_type=None,
            contextual_value=None,
            op_intermediate_type=None,
            resolved_symbol=None,
        )

    def visit_AstFor(self, node: AstFor, state: CompileState):
        assert isinstance(node.range, AstRange), node.range

        # transform this:

        # for <node.loop_var> in <node.range>:
        #     <node.body>

        # to:

        # 1 <node.loop_var>: LoopVarType = <node.range.lower_bound>
        # OR (depending on whether redeclaring or not)
        # 1 <node.loop_var> = <node.range.lower_bound>
        # 2 $upper_bound_var: LoopVarType = <node.range.upper_bound>
        # 3 while <node.loop_var> < $upper_bound_var:
        #      <node.body>
        #      <node.loop_var> = <node.loop_var> + 1

        loop_info = state.for_loops[node]

        # 1
        initialize_loop_var = self.initialize_loop_var(state, node, loop_info)
        # 2
        declare_upper_bound_var = self.declare_upper_bound_var(state, node, loop_info)
        # 3
        while_loop: AstWhile = self.while_loop(state, node, loop_info)

        # this is the first and so far only piece of code in the compiler itself written by AI
        # Update any break/continue statements in the body to point to new while loop
        # instead of the original for loop
        for key, value in list(state.enclosing_loops.items()):
            if value == node:  # If a break/continue was pointing to our for loop
                state.enclosing_loops[key] = (
                    while_loop  # Point it to while loop instead
                )

        state.desugared_for_loops[while_loop] = node

        # turn one node into three
        return [initialize_loop_var, declare_upper_bound_var, while_loop]


class DesugarDefaultArgs(Transformer):
    """
    Desugars function calls with named or missing arguments by:
    1. Reordering named arguments to positional order
    2. Filling in default values for missing arguments
    3. Converting FppValue defaults (from builtins) to AstNumber/AstBoolean nodes

    For example, if we have:
        def foo(a: U8, b: U8 = 5, c: U8 = 10):
            pass
        foo(c=15, a=1)

    This becomes:
        foo(1, 5, 15)

    Note: The type coercion for default values is handled during semantic analysis
    in PickTypesAndResolveAttrsAndItems.visit_AstDef. By the time this desugaring
    runs, contextual_types already has the correct coerced types for default
    value expressions.
    """

    def _fpp_value_to_ast(self, value: FppValue, meta: Meta, state: CompileState) -> Ast:
        """Convert an FppValue (from builtin default) to an AST literal node."""
        if isinstance(value, BoolValue):
            node = AstBoolean(meta=meta, value=value.val)
        elif isinstance(value, IntegerValue):
            node = AstNumber(meta=meta, value=value.val)
        else:
            assert False, f"Unsupported FppValue type for default arg: {type(value)}"

        # Register the new node in state so codegen can find its type/value
        state.synthesized_types[node] = type(value)
        state.contextual_types[node] = type(value)
        state.contextual_values[node] = value
        return node

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        # Get the resolved arguments from semantic analysis.
        # This list is already in positional order with defaults filled in.
        resolved_args = state.resolved_func_args.get(node)
        assert resolved_args is not None, (
            f"No resolved args for function call {node}. "
            f"This should have been set by PickTypesAndResolveAttrsAndItems."
        )

        # Convert any FppValue defaults to AST nodes
        desugared_args = []
        for arg in resolved_args:
            if isinstance(arg, FppValue):
                desugared_args.append(self._fpp_value_to_ast(arg, node.meta, state))
            else:
                desugared_args.append(arg)

        # Update the node's args with the desugared arguments
        node.args = desugared_args

        return node



from __future__ import annotations
from datetime import datetime, timezone
from decimal import Decimal
import decimal
import struct
from numbers import Number
from typing import Union

from fpy.error import CompileError
from fpy.types import (
    ARBITRARY_PRECISION_TYPES,
    SIGNED_INTEGER_TYPES,
    SPECIFIC_NUMERIC_TYPES,
    UNSIGNED_INTEGER_TYPES,
    BuiltinSymbol,
    CompileState,
    FieldSymbol,
    ForLoopAnalysis,
    FppType,
    CallableSymbol,
    CastSymbol,
    FpyFloatValue,
    FunctionSymbol,
    Symbol,
    SymbolTable,
    TypeCtorSymbol,
    VariableSymbol,
    FpyIntegerValue,
    FpyStringValue,
    NothingValue,
    RangeValue,
    TopDownVisitor,
    Visitor,
    is_instance_compat,
    lookup_symbol,
    typename,
)

# In Python 3.10+, the `|` operator creates a `types.UnionType`.
# We need to handle this for forward compatibility, but it won't exist in 3.9.
try:
    from types import UnionType

    UNION_TYPES = (Union, UnionType)
except ImportError:
    UNION_TYPES = (Union,)

from fpy.bytecode.directives import (
    BOOLEAN_OPERATORS,
    NUMERIC_OPERATORS,
    ArrayIndexType,
    LoopVarType,
    BinaryStackOp,
    UnaryStackOp,
)
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.templates.prm_template import PrmTemplate
from fprime_gds.common.models.serialize.time_type import TimeType as TimeValue
from fprime_gds.common.models.serialize.type_base import ValueType
from fprime_gds.common.models.serialize.serializable_type import (
    SerializableType as StructValue,
)
from fprime_gds.common.models.serialize.array_type import ArrayType as ArrayValue
from fprime_gds.common.models.serialize.type_exceptions import TypeException
from fprime_gds.common.models.serialize.numerical_types import (
    U8Type as U8Value,
    U16Type as U16Value,
    U32Type as U32Value,
    U64Type as U64Value,
    I64Type as I64Value,
    F64Type as F64Value,
    FloatType as FloatValue,
    IntegerType as IntegerValue,
    NumericalType as NumericalValue,
)
from fprime_gds.common.models.serialize.string_type import StringType as StringValue
from fprime_gds.common.models.serialize.bool_type import BoolType as BoolValue
from fpy.syntax import (
    AstAssert,
    AstBinaryOp,
    AstStmtList,
    AstBoolean,
    AstBreak,
    AstContinue,
    AstDef,
    AstElif,
    AstExpr,
    AstFor,
    AstMemberAccess,
    AstIndexExpr,
    AstTypeExpr,
    AstNamedArgument,
    AstNumber,
    AstPass,
    AstRange,
    AstReference,
    AstReturn,
    AstBlock,
    AstStmt,
    AstStmtWithExpr,
    AstString,
    Ast,
    AstBlock,
    AstLiteral,
    AstIf,
    AstAssign,
    AstFuncCall,
    AstUnaryOp,
    AstVar,
    AstWhile,
)
from fprime_gds.common.models.serialize.type_base import BaseType as FppValue


class AssignIds(TopDownVisitor):
    """assigns a unique id to each node to allow it to be indexed in a dict"""

    def visit_default(self, node, state: CompileState):
        node.id = state.next_node_id
        state.next_node_id += 1


class SetLocalScope(Visitor):
    def __init__(self, scope: SymbolTable):
        super().__init__()
        self.scope = scope

    def visit_default(self, node: Ast, state: CompileState):
        state.local_scopes[node] = self.scope


class AssignLocalScopes(TopDownVisitor):

    def visit_AstBlock(self, node: AstBlock, state: CompileState):
        if node is not state.root:
            # only handle the root node this way
            return
        # make a new scope
        scope = SymbolTable()
        state.scope_parents[scope] = None
        # TODO ask rob there must be a better way to do this, that isn't as slow
        SetLocalScope(scope).run(node, state)

    def visit_AstDef(self, node: AstDef, state: CompileState):
        parent_scope = state.local_scopes[node]
        # make a new scope for the function body
        scope = SymbolTable()
        state.scope_parents[scope] = parent_scope

        # The function body gets the new scope
        SetLocalScope(scope).run(node.body, state)

        # Parameter names and type annotations are in the body scope
        # (they're declared inside the function)
        if node.parameters is not None:
            for arg_name_var, arg_type_expr, default_value in node.parameters:
                state.local_scopes[arg_name_var] = scope
                state.local_scopes[arg_type_expr] = scope
                # Default values stay in parent scope - they're evaluated at call site,
                # not inside the function body
                if default_value is not None:
                    SetLocalScope(parent_scope).run(default_value, state)

        # Return type annotation is in parent scope (it references types visible at def site)
        if node.return_type is not None:
            state.local_scopes[node.return_type] = parent_scope

        # The def node itself is in the parent scope
        state.local_scopes[node] = parent_scope
        # Function name is in parent scope
        state.local_scopes[node.name] = parent_scope


class CreateVariablesAndFuncs(TopDownVisitor):
    """finds all variable declarations and adds them to the variable scope"""

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        if not is_instance_compat(node.lhs, AstReference):
            # trying to assign a value to some complex expression like (1 + 1) = 2
            state.err("Invalid assignment", node.lhs)
            return

        if not is_instance_compat(node.lhs, AstVar):
            # assigning to a member or array element. don't need to make a new variable,
            # space already exists
            if node.type_ann is not None:
                # type annotation on a field assignment... it already has a type!
                state.err("Cannot specify a type annotation for a field", node.type_ann)
                return
            # otherwise we good
            return

        if node.type_ann is not None:
            # new variable declaration
            # make sure it isn't defined in this scope
            # TODO shadowing check
            existing_local = state.local_scopes[node].get(node.lhs.var)
            if existing_local is not None:
                # redeclaring an existing variable
                state.err(f"'{node.lhs.var}' has already been declared", node)
                return
            # okay, declare the var
            # Check if we're in the root (global) scope
            current_scope = state.local_scopes[node]
            is_global = state.scope_parents.get(current_scope) is None
            var = VariableSymbol(node.lhs.var, node.type_ann, node, is_global=is_global)
            # new var. put it in the table under this scope
            state.local_scopes[node][node.lhs.var] = var
        else:
            # otherwise, it's a reference to an existing var
            sym = lookup_symbol(node, node.lhs.var, state)
            if sym is None:
                # unable to find this symbol
                state.err(
                    f"'{node.lhs.var}' has not been declared",
                    node.lhs,
                )
                return
            # okay, we were able to resolve it

    def visit_AstFor(self, node: AstFor, state: CompileState):
        # for loops have an implicit loop variable that they can declare
        # if it isn't already declared in the local scope
        loop_var = state.local_scopes[node].get(node.loop_var.var)

        reuse_existing_loop_var = False
        if loop_var is not None:
            # this is okay as long as the variable is of the same type

            # what follows is a bit of a hack
            # there are two cases: either loop_var has been declared before but we only know the type expr (if it was an AstAssign decl)
            # or loop_var has been declared before and we only know the type, but have no type expr

            # case 1 is easy, just check the type == LoopVarType
            # case 2 is harder, we have to check if the type expr is an AstTypeExpr with a single part
            # that matches the canonical name of the LoopVarType

            # the alternative to this is that we do some primitive type resolution in the same pass as variable creation
            # i'm doing this hack because we're going to switch to type inference for variables later and that will make this go away

            if (loop_var.type_ref is None and loop_var.type != LoopVarType) or (
                loop_var.type is None
                and not (
                    isinstance(loop_var.type_ref, AstTypeExpr)
                    and loop_var.type_ref.parts == [LoopVarType.get_canonical_name()]
                )
            ):
                state.err(
                    f"'{node.loop_var.var}' has already been declared as a type other than {typename(LoopVarType)}",
                    node,
                )
                return
            reuse_existing_loop_var = True
        else:
            # new var. put it in the table under this scope
            loop_var = VariableSymbol(node.loop_var.var, None, node, LoopVarType)
            state.local_scopes[node][loop_var.name] = loop_var

        # each loop also declares an implicit ub variable
        # type of ub var is same as loop var type
        upper_bound_var = VariableSymbol(
            state.new_anonymous_variable_name(), None, node, LoopVarType
        )
        state.local_scopes[node][upper_bound_var.name] = upper_bound_var
        analysis = ForLoopAnalysis(loop_var, upper_bound_var, reuse_existing_loop_var)
        state.for_loops[node] = analysis

    def visit_AstDef(self, node: AstDef, state: CompileState):
        existing_func = state.local_scopes[node].get(node.name.var)
        if existing_func is not None:
            state.err(f"'{node.name.var}' has already been declared", node.name)
            return

        func = FunctionSymbol(
            # we know the name
            node.name.var,
            # we don't know the return type yet
            return_type=None,
            # we don't know the arg types yet
            args=None,
            definition=node,
        )

        state.local_scopes[node][func.name] = func

        if node.parameters is None:
            # no arguments
            return

        # Check that default arguments come after non-default arguments
        seen_default = False
        for arg in node.parameters:
            arg_name_var, arg_type_expr, default_value = arg
            if default_value is not None:
                seen_default = True
            elif seen_default:
                # Non-default argument after default argument
                state.err(
                    f"Non-default argument '{arg_name_var.var}' follows default argument",
                    arg_name_var,
                )
                return

        for arg in node.parameters:
            arg_name_var, arg_type_expr, default_value = arg
            existing_local = state.local_scopes[node.body].get(arg_name_var.var)
            if existing_local is not None:
                # redeclaring an existing variable
                state.err(
                    f"'{arg_name_var.var}' has already been declared", arg_name_var
                )
                return
            arg_var = VariableSymbol(arg_name_var.var, arg_type_expr, node)
            state.local_scopes[node.body][arg_name_var.var] = arg_var


class SetEnclosingLoops(Visitor):
    """sets or clears the enclosing_loop dict of any break/continue it finds"""

    def __init__(self, loop: Union[AstFor, AstWhile]):
        """if loop is None, remove the visited break/continue from enclosing_loop dict"""
        super().__init__()
        self.loop = loop

    def visit_AstBreak_AstContinue(
        self, node: Union[AstBreak, AstContinue], state: CompileState
    ):
        if self.loop is None:
            del state.enclosing_loops[node]
        else:
            state.enclosing_loops[node] = self.loop


class CheckBreakAndContinueInLoop(TopDownVisitor):
    def visit_AstFor_AstWhile(self, node: Union[AstFor, AstWhile], state: CompileState):
        SetEnclosingLoops(node).run(node.body, state)

    def visit_AstBreak_AstContinue(
        self, node: Union[AstBreak, AstContinue], state: CompileState
    ):
        if node not in state.enclosing_loops:
            state.err("Cannot break/continue outside of a loop", node)
            return

    def visit_AstDef(self, node: AstDef, state: CompileState):
        # going inside of a func def "resets" our loop context. this prevents the following scenario:
        # for x in 0..2:
        #     def test():
        #         break

        # in this case, the break should fail to compile
        SetEnclosingLoops(None).run(node.body, state)


class SetEnclosingFunction(Visitor):
    def __init__(self, func: AstDef):
        super().__init__()
        self.func = func

    def visit_AstReturn(self, node: AstReturn, state: CompileState):
        state.enclosing_funcs[node] = self.func


class CheckReturnInFunc(TopDownVisitor):
    def visit_AstDef(self, node: AstDef, state: CompileState):
        SetEnclosingFunction(node).run(node.body, state)

    def visit_AstReturn(self, node: AstReturn, state: CompileState):
        if node not in state.enclosing_funcs:
            state.err("Cannot return outside of a function", node)
            return


class ResolveTypeNames(TopDownVisitor):
    """
    Resolves type annotations (AstTypeExpr) to actual types.
    This runs before ResolveVars so that variable types are known
    when we start resolving references.
    """

    def resolve_type_name(
        self, node: AstTypeExpr, state: CompileState
    ) -> type | None:
        """
        Fully resolves a type name to the actual type.
        Returns None if the type could not be resolved (error already reported).
        """
        # Start from the first part
        sym = state.types.get(node.parts[0])
        if sym is None:
            state.err("Unknown type", node)
            return None

        # Walk through the remaining parts
        for part in node.parts[1:]:
            if not is_instance_compat(sym, dict):
                state.err("Unknown type", node)
                return None
            sym = sym.get(part)
            if sym is None:
                state.err("Unknown type", node)
                return None

        if not is_instance_compat(sym, type):
            state.err("Unknown type", node)
            return None

        state.resolved_symbols[node] = sym
        return sym

    def visit_AstDef(self, node: AstDef, state: CompileState):
        # Get the function that was created in CreateVariablesAndFuncs
        func = state.local_scopes[node].get(node.name.var)
        assert is_instance_compat(func, FunctionSymbol), func

        # Resolve return type
        if node.return_type is not None:
            return_type = self.resolve_type_name(node.return_type, state)
            if return_type is None:
                return
            func.return_type = return_type
        else:
            func.return_type = NothingValue

        # Resolve parameter types
        args = []
        if node.parameters is not None:
            for arg_name_var, arg_type_expr, default_value in node.parameters:
                arg_type = self.resolve_type_name(arg_type_expr, state)
                if arg_type is None:
                    return

                # Get the variable that was created for this parameter
                arg_var = state.local_scopes[node.body].get(arg_name_var.var)
                assert is_instance_compat(arg_var, VariableSymbol), arg_var
                arg_var.type = arg_type
                args.append((arg_name_var.var, arg_type, default_value))

        func.args = args

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        if node.type_ann is None:
            return

        var_type = self.resolve_type_name(node.type_ann, state)
        if var_type is None:
            return

        # Get the variable - it should be in the local scope
        if not is_instance_compat(node.lhs, AstVar):
            # Type annotations only make sense on simple variable assignments
            state.err("Type annotation can only be on simple variable assignment", node)
            return

        var = state.local_scopes[node].get(node.lhs.var)
        assert is_instance_compat(var, VariableSymbol), f"Variable {node.lhs.var} should have been created by CreateVariablesAndFuncs"

        var.type = var_type


class ResolveVars(TopDownVisitor):
    """
    Resolves all variable references (AstVar) in local and global scopes.
    Also fully resolves function references for function calls.
    Types are already resolved by ResolveTypeNames before this pass.
    """

    def resolve_local_name(
        self, node: AstVar, state: CompileState
    ) -> Symbol | None:
        """resolves a name in local scope only. return None if could not be resolved"""
        local_scope = state.local_scopes[node]
        sym = None
        while local_scope is not None and sym is None:
            sym = local_scope.get(node.var)
            local_scope = state.scope_parents[local_scope]

        if sym is not None:
            state.resolved_symbols[node] = sym
        return sym

    def try_resolve_root_ref(
        self,
        node: Ast,
        global_scope: SymbolTable,
        global_scope_name: str,
        state: CompileState,
        search_local_scope: bool = True,
    ) -> bool:
        """
        recursively tries to resolve the root of a reference. if any node in the chain is not a reference, return True.
        Otherwise recurse up the sym until you find the root, and try resolving
        it in the local scope, and then in the given global scope. Return False if couldn't be resolved in local and global
        """
        if not is_instance_compat(node, AstReference):
            # not a reference, nothing to resolve
            return True

        if not is_instance_compat(node, AstVar):
            # it is a reference but it's not a var
            # recurse until we find the var
            return self.try_resolve_root_ref(
                node.parent, global_scope, global_scope_name, state
            )

        # okay now we have a var
        # see if it's something defined in the script
        sym = None
        if search_local_scope:
            sym = self.resolve_local_name(node, state)

        if sym is None:
            # unable to find this symbol in the hierarchy of local scopes
            # look it up in the global scope
            sym = global_scope.get(node.var)

        if sym is None:
            state.err(f"Unknown {global_scope_name}", node)
            return False

        state.resolved_symbols[node] = sym
        return True

    def finish_resolving_func(
        self,
        node: Ast,
        state: CompileState,
    ) -> CallableSymbol | None:
        """
        Finishes resolving a function reference (AstVar/AstMemberAccess chain) to a callable.
        The root AstVar should already be resolved by try_resolve_root_ref.
        Returns None if the reference could not be resolved (error already reported).
        """

        def resolve(n: Ast, expect_callable: bool) -> CallableSymbol | dict | None:
            """
            Recursively resolve the reference chain.
            expect_callable=True for the final node (must be CallableSymbol),
            expect_callable=False for intermediate nodes (must be dict/namespace).
            """
            if not is_instance_compat(n, (AstVar, AstMemberAccess)):
                state.err("Unknown function", n)
                return None

            if is_instance_compat(n, AstVar):
                sym = state.resolved_symbols.get(n)
                if sym is None:
                    state.err("Unknown function", n)
                    return None
                expected_type = CallableSymbol if expect_callable else dict
                if not is_instance_compat(sym, expected_type):
                    state.err("Unknown function", n)
                    return None
                return sym

            # It's an AstMemberAccess - resolve the parent first (always expecting a namespace)
            parent_scope = resolve(n.parent, expect_callable=False)
            if parent_scope is None:
                return None

            sym = parent_scope.get(n.attr)
            if sym is None:
                state.err("Unknown function", n)
                return None

            expected_type = CallableSymbol if expect_callable else dict
            if not is_instance_compat(sym, expected_type):
                state.err("Unknown function", n)
                return None

            state.resolved_symbols[n] = sym
            return sym

        return resolve(node, expect_callable=True)

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        # First resolve the root of the function reference
        if not self.try_resolve_root_ref(node.func, state.callables, "function", state):
            return

        # Then finish resolving the full chain to get the callable
        if not self.finish_resolving_func(node.func, state):
            return

        for arg in node.args if node.args is not None else []:
            # Handle both positional args (AstExpr) and named args (AstNamedArgument)
            if is_instance_compat(arg, AstNamedArgument):
                # For named arguments, resolve the value expression
                if not self.try_resolve_root_ref(
                    arg.value, state.runtime_values, "value", state
                ):
                    return
            else:
                # arg value refs must have values at runtime
                if not self.try_resolve_root_ref(
                    arg, state.runtime_values, "value", state
                ):
                    return

    def visit_AstIf_AstElif(self, node: Union[AstIf, AstElif], state: CompileState):
        # if condition expr refs must be "runtime values" (tlm/prm/const/etc)
        if not self.try_resolve_root_ref(
            node.condition, state.runtime_values, "value", state
        ):
            return

    def visit_AstBinaryOp(self, node: AstBinaryOp, state: CompileState):
        # lhs/rhs side of stack op, if they are refs, must be refs to "runtime vals"
        if not self.try_resolve_root_ref(
            node.lhs, state.runtime_values, "value", state
        ):
            return
        if not self.try_resolve_root_ref(
            node.rhs, state.runtime_values, "value", state
        ):
            return

    def visit_AstUnaryOp(self, node: AstUnaryOp, state: CompileState):
        if not self.try_resolve_root_ref(
            node.val, state.runtime_values, "value", state
        ):
            return

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        if not self.try_resolve_root_ref(
            node.lhs, state.runtime_values, "value", state
        ):
            return

        # Type annotation is resolved by ResolveTypeNames pass

        if not self.try_resolve_root_ref(
            node.rhs, state.runtime_values, "value", state
        ):
            return

    def visit_AstFor(self, node: AstFor, state: CompileState):
        if not self.try_resolve_root_ref(
            node.loop_var, state.runtime_values, "value", state
        ):
            return

        # this really shouldn't be possible to be a var right now
        # but this is future proof
        if not self.try_resolve_root_ref(
            node.range, state.runtime_values, "value", state
        ):
            return

    def visit_AstWhile(self, node: AstWhile, state: CompileState):
        if not self.try_resolve_root_ref(
            node.condition, state.runtime_values, "value", state
        ):
            return

    def visit_AstAssert(self, node: AstAssert, state: CompileState):
        if not self.try_resolve_root_ref(
            node.condition, state.runtime_values, "value", state
        ):
            return
        if node.exit_code is not None:
            if not self.try_resolve_root_ref(
                node.exit_code, state.runtime_values, "value", state
            ):
                return

    def visit_AstVar(self, node: AstVar, state: CompileState):
        # if this var isn't resolved, that means it wasn't used in
        # any of the other places that it could. it must be alone on its line
        # this is a strange choice by the dev but rule of thumb says start
        # by allowing everything, then add warnings later
        if node in state.resolved_symbols:
            return

        if self.resolve_local_name(node, state) is None:
            state.err("Unknown symbol", node)
            return

    def visit_AstIndexExpr(self, node: AstIndexExpr, state: CompileState):
        if not self.try_resolve_root_ref(
            node.item, state.runtime_values, "value", state
        ):
            return

    def visit_AstRange(self, node: AstRange, state: CompileState):
        if not self.try_resolve_root_ref(
            node.lower_bound, state.runtime_values, "value", state
        ):
            return
        if not self.try_resolve_root_ref(
            node.upper_bound, state.runtime_values, "value", state
        ):
            return

    def visit_AstDef(self, node: AstDef, state: CompileState):
        if not self.try_resolve_root_ref(node.name, state.callables, "function", state):
            return

        # Return type and parameter types are resolved by ResolveTypeNames pass

        if node.parameters is not None:
            for arg_name_var, arg_type_expr, default_value in node.parameters:
                if not self.try_resolve_root_ref(
                    arg_name_var, state.runtime_values, "value", state
                ):
                    return

                # Type is resolved by ResolveTypeNames pass

                # Resolve default value if present
                if default_value is not None:
                    if not self.try_resolve_root_ref(
                        default_value, state.runtime_values, "value", state
                    ):
                        return

    def visit_AstReturn(self, node: AstReturn, state: CompileState):
        if not self.try_resolve_root_ref(
            node.value, state.runtime_values, "value", state
        ):
            return

    def visit_AstLiteral_AstMemberAccess(
        self, node: Union[AstLiteral, AstMemberAccess], state: CompileState
    ):
        # don't need to do anything for literals or getattr, but just have this here for completion's sake
        # the reason we don't need to do anything for getattr is because the point of this
        # pass is explicitly not to resolve getattr, just resolve vars (and types), not general
        # getattr
        # we don't do this now because we need to resolve all expr types first
        # before we can figure out whether a getattr is correct, and this pass lets
        # us now do that (i.e. after this pass, we have enough info about the program
        # that we can decide types of any expr)
        pass

    def visit_default(self, node, state):
        # coding error, missed an expr
        assert not is_instance_compat(node, AstStmtWithExpr), node


class CheckUseBeforeDeclare(TopDownVisitor):
    """
    Checks that variables are not used before they are declared.
    Handles both regular variable assignments (AstAssign) and for loop variables (AstFor).
    
    Uses TopDownVisitor because for loops need the loop variable to be declared
    before visiting the body. For assignments, we manually check the RHS before
    marking the variable as declared.
    """

    def __init__(self):
        super().__init__()
        self.currently_declared_vars: list[VariableSymbol] = []

    def visit_AstFor(self, node: AstFor, state: CompileState):
        var = state.resolved_symbols[node.loop_var]
        # Check that the loop var isn't referenced in the range (before it's declared)
        EnsureVariableNotReferenced(var).run(node.range, state)
        # Now mark it as declared for the body
        self.currently_declared_vars.append(var)

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        if not is_instance_compat(node.lhs, AstVar):
            # definitely not a declaration, it's a field assignment
            return

        var = state.resolved_symbols[node.lhs]

        if var is None or var.declaration != node:
            # either not declared in this scope, or this is not a
            # declaration of this var
            return

        # Before marking as declared, check that the variable isn't used in its own RHS
        EnsureVariableNotReferenced(var).run(node.rhs, state)

        # Now mark this variable as declared
        self.currently_declared_vars.append(var)

    def visit_AstVar(self, node: AstVar, state: CompileState):
        sym = state.resolved_symbols[node]
        if not is_instance_compat(sym, VariableSymbol):
            # not a variable, might be a type name or smth
            return

        if is_instance_compat(sym.declaration, AstDef):
            # function parameters - no use-before-declare check needed
            # this is because if it's in scope, it's declared, as its
            # "declaration" is the start of the scope
            return
        if (
            is_instance_compat(sym.declaration, AstAssign)
            and sym.declaration.lhs == node
        ):
            # this is the declaring reference for an assignment
            return
        if (
            is_instance_compat(sym.declaration, AstFor)
            and sym.declaration.loop_var == node
        ):
            # this is the declaring reference for a for loop variable
            return

        if sym not in self.currently_declared_vars:
            state.err(f"'{node.var}' used before declared", node)
            return


class EnsureVariableNotReferenced(Visitor):
    def __init__(self, var: VariableSymbol):
        super().__init__()
        self.var = var

    def visit_AstVar(self, node: AstVar, state: CompileState):
        sym = state.resolved_symbols[node]
        if sym == self.var:
            state.err(f"'{node.var}' used before declared", node)
            return


class PickTypesAndResolveAttrsAndItems(Visitor):

    def coerce_expr_type(
        self, node: AstExpr, type: FppType, state: CompileState
    ) -> bool:
        unconverted_type = state.synthesized_types[node]
        # make sure it isn't already being coerced
        assert unconverted_type == state.contextual_types[node], (
            unconverted_type,
            state.contextual_types[node],
        )
        if self.can_coerce_type(unconverted_type, type):
            state.contextual_types[node] = type
            return True
        state.err(
            f"Expected {typename(type)}, found {typename(unconverted_type)}", node
        )
        return False

    def can_coerce_type(self, from_type: FppType, to_type: FppType) -> bool:
        """return True if the type coercion rules allow from_type to be implicitly converted to to_type"""
        if from_type == to_type:
            # no coercion necessary
            return True
        if from_type == FpyStringValue and issubclass(to_type, StringValue):
            # we can convert the literal String type to any string type
            return True
        if not issubclass(from_type, NumericalValue) or not issubclass(
            to_type, NumericalValue
        ):
            # if one of the src or dest aren't numerical, we can't coerce
            return False

        # now we must answer:
        # are all values of from_type representable in the destination type?

        # if going from float to integer, definitely not
        if issubclass(from_type, FloatValue) and issubclass(to_type, IntegerValue):
            return False

        # in general: if either src or dest is one of our FpyXYZValue types, which are
        # arb precision, we allow this coercion.
        # it's easy to argue we should allow converting to arb precision. but why would
        # we allow arb precision to go to an 8 bit type, e.g.?
        # we have a big advantage: the arb precision types are only used for constants. that
        # means we actually know what the value is, so we can actually check!
        # however, we won't perform that check here. That will happen later in the
        # const_convert_type func in the CalcConstExprValues
        # for now, we will let the compilation proceed if either side is arb precision

        if (
            from_type in ARBITRARY_PRECISION_TYPES
            or to_type in ARBITRARY_PRECISION_TYPES
        ):
            return True

        # otherwise, both src and dest have finite bits

        # if we currently have a float
        if issubclass(from_type, FloatValue):
            # the dest must be a float and must be >= width
            return (
                issubclass(to_type, FloatValue)
                and to_type.get_bits() >= from_type.get_bits()
            )

        # otherwise must be an int
        assert issubclass(from_type, IntegerValue)
        # int to float is allowed in any case.
        # this is the big exception to our rule about full representation. this can cause loss of precision
        # for large integer values
        if issubclass(to_type, FloatValue):
            return True

        # the dest must be an int with the same signedness and >= width
        from_unsigned = from_type in UNSIGNED_INTEGER_TYPES
        to_unsigned = to_type in UNSIGNED_INTEGER_TYPES
        return (
            from_unsigned == to_unsigned and to_type.get_bits() >= from_type.get_bits()
        )

    def pick_intermediate_type(
        self, arg_types: list[FppType], op: BinaryStackOp | UnaryStackOp
    ) -> FppType:
        """return the intermediate type that all arguments should be converted to for the given operator"""

        if op in BOOLEAN_OPERATORS:
            return BoolValue

        non_numeric = any(not issubclass(t, NumericalValue) for t in arg_types)

        if (op == BinaryStackOp.EQUAL or op == BinaryStackOp.NOT_EQUAL) and non_numeric:
            # comparison of complex types (structs/strings/arrays/enum consts)
            if len(set(arg_types)) != 1:
                # can only compare equality between the same types
                return None
            return arg_types[0]

        # all other cases require that arguments are numeric
        if non_numeric:
            return None

        # we split this algo up into two stages: picking the type category (float, uint or int), and picking the type bitwidth

        # pick the type category:
        type_category = None
        if op == BinaryStackOp.DIVIDE or op == BinaryStackOp.EXPONENT:
            # always do true division and exponentiation over floats, python style
            # this is because, for the given op, even with integer inputs, we might get
            # float outputs
            type_category = "float"
            # TODO problem: this means that if we do I32 / F32, it doesn't work b/c we can't convert I32 to F32
        elif any(issubclass(t, FloatValue) for t in arg_types):
            # otherwise if any args are floats, use float
            type_category = "float"
        elif any(t in UNSIGNED_INTEGER_TYPES for t in arg_types):
            # otherwise if any args are unsigned, use unsigned
            type_category = "uint"
        else:
            # otherwise use signed int
            type_category = "int"

        # pick the bitwidth
        # we only use the arb precision types for constants, so if theyre all arb precision, they're consts
        constants = all(t in ARBITRARY_PRECISION_TYPES for t in arg_types)

        if constants:
            # we can constant fold this, so use infinite bitwidth
            if type_category == "float":
                return FpyFloatValue
            assert type_category == "int" or type_category == "uint"
            return FpyIntegerValue

        # can't const fold
        if type_category == "float":
            return F64Value
        if type_category == "uint":
            return U64Value
        assert type_category == "int"
        return I64Value

    def is_type_constant_size(self, type: FppType) -> bool:
        """return true if the type is statically sized"""
        if issubclass(type, StringValue):
            return False

        if issubclass(type, ArrayValue):
            return self.is_type_constant_size(type.MEMBER_TYPE)

        if issubclass(type, StructValue):
            for _, arg_type, _, _ in type.MEMBER_LIST:
                if not self.is_type_constant_size(arg_type):
                    return False
            return True

        return True

    def get_members(
        self, node: Ast, parent_type: FppType, state: CompileState
    ) -> list[tuple[str, FppType]] | None:
        if not issubclass(parent_type, (StructValue, TimeValue)):
            return {}

        if not self.is_type_constant_size(parent_type):
            state.err(
                f"{parent_type} has non-constant sized members, cannot access members",
                node,
            )
            return None

        member_list: list[tuple[str, FppType]] = None
        if issubclass(parent_type, StructValue):
            member_list = [t[0:2] for t in parent_type.MEMBER_LIST]
        else:
            # if it is a time type, there are some "implied" members
            member_list = []
            member_list.append(("time_base", U16Value))
            member_list.append(("time_context", U8Value))
            member_list.append(("seconds", U32Value))
            member_list.append(("useconds", U32Value))
        return member_list

    def get_sym_type(self, sym: Symbol) -> FppType:
        """returns the fprime type of the sym, if it were to be evaluated as an expression"""
        if isinstance(sym, ChTemplate):
            result_type = sym.ch_type_obj
        elif isinstance(sym, PrmTemplate):
            result_type = sym.prm_type_obj
        elif isinstance(sym, FppValue):
            # constant value
            result_type = type(sym)
        elif isinstance(sym, CallableSymbol):
            # a reference to a callable isn't a type in and of itself
            # it has a return type but you have to call it (with an AstFuncCall)
            # consider making a separate "reference" type
            result_type = NothingValue
        elif isinstance(sym, VariableSymbol):
            result_type = sym.type
        elif isinstance(sym, type):
            # a reference to a type doesn't have a value, and so doesn't have a type,
            # in and of itself. if this were a function call to the type's ctor then
            # it would have a value and thus a type
            result_type = NothingValue
        elif isinstance(sym, FieldSymbol):
            result_type = sym.type
        elif isinstance(sym, dict):
            # reference to a scope. scopes don't have values
            result_type = NothingValue
        else:
            assert False, sym

        return result_type

    def visit_AstMemberAccess(self, node: AstMemberAccess, state: CompileState):
        parent_sym = state.resolved_symbols.get(node.parent)

        if is_instance_compat(parent_sym, (type, CallableSymbol)):
            state.err("Unknown attribute", node)
            return

        sym = None
        if is_instance_compat(parent_sym, dict):
            # getattr of a namespace
            # parent won't actually have a type
            sym = parent_sym.get(node.attr)
            if sym is None:
                state.err("Unknown attribute", node)
                return
            # GetAttr should never resolve to a lexical variable; variables are accessed directly
            assert not is_instance_compat(
                sym, VariableSymbol
            ), "Field resolution unexpectedly found a local variable"
        else:
            # in all other cases, parent has at least some sort of type
            # sym may be None (if parent is some complex expr), or it may be
            # a tlm chan or var or etc...
            # it may or may not have a compile time value, but it definitely has a type
            parent_type = state.synthesized_types[node.parent]

            # field symbols store their "base symbol", which is the first non-field-symbol parent of
            # the field symbol. this lets you easily check what actual underlying thing (tlm chan, variable, prm)
            # you're talking about a field of
            base_sym = (
                parent_sym
                if not is_instance_compat(parent_sym, FieldSymbol)
                else parent_sym.base_sym
            )
            # we also calculate a "base offset" wrt. the start of the base_sym type, so you
            # can easily pick out this field from a value of the base sym type
            base_offset = (
                0
                if not is_instance_compat(parent_sym, FieldSymbol)
                else parent_sym.base_offset
            )

            member_list = self.get_members(node, parent_type, state)
            if member_list is None:
                return

            offset = 0
            for arg_name, arg_type in member_list:
                if arg_name == node.attr:
                    sym = FieldSymbol(
                        is_struct_member=True,
                        parent_expr=node.parent,
                        type=arg_type,
                        base_sym=base_sym,
                        local_offset=offset,
                        base_offset=base_offset,
                        name=arg_name,
                    )
                    break
                offset += arg_type.getMaxSize()
                base_offset += arg_type.getMaxSize()

        if sym is None:
            state.err(
                f"{typename(parent_type)} has no member named {node.attr}",
                node,
            )
            return

        sym_type = self.get_sym_type(sym)

        state.resolved_symbols[node] = sym
        state.synthesized_types[node] = sym_type
        state.contextual_types[node] = sym_type

    def visit_AstIndexExpr(self, node: AstIndexExpr, state: CompileState):
        parent_sym = state.resolved_symbols.get(node.parent)

        if is_instance_compat(parent_sym, (type, CallableSymbol, dict)):
            state.err("Unknown item", node)
            return

        # otherwise, we should definitely have a well-defined type for our parent expr

        parent_type = state.synthesized_types[node.parent]

        if not self.is_type_constant_size(parent_type):
            state.err(
                f"{typename(parent_type)} has non-constant sized members, cannot access items",
                node,
            )
            return

        if not issubclass(parent_type, ArrayValue):
            state.err(f"{typename(parent_type)} is not an array", node)
            return

        # coerce the index expression to array index type
        if not self.coerce_expr_type(node.item, ArrayIndexType, state):
            return

        base_sym = (
            parent_sym
            if not is_instance_compat(parent_sym, FieldSymbol)
            else parent_sym.base_sym
        )

        sym = FieldSymbol(
            is_array_element=True,
            parent_expr=node.parent,
            type=parent_type.MEMBER_TYPE,
            base_sym=base_sym,
            idx_expr=node.item,
        )

        state.resolved_symbols[node] = sym
        state.synthesized_types[node] = parent_type.MEMBER_TYPE
        state.contextual_types[node] = parent_type.MEMBER_TYPE

    def visit_AstVar(self, node: AstVar, state: CompileState):
        # already been resolved by SetScopes pass
        sym = state.resolved_symbols[node]
        if sym is None:
            return
        sym_type = self.get_sym_type(sym)

        state.synthesized_types[node] = sym_type
        state.contextual_types[node] = sym_type

    def visit_AstNumber(self, node: AstNumber, state: CompileState):
        # give a best guess as to the final type of this node. we don't actually know
        # its bitwidth or signedness yet
        if is_instance_compat(node.value, Decimal):
            result_type = FpyFloatValue
        else:
            result_type = FpyIntegerValue

        state.synthesized_types[node] = result_type
        state.contextual_types[node] = result_type

    def visit_AstBinaryOp(self, node: AstBinaryOp, state: CompileState):
        lhs_type = state.synthesized_types[node.lhs]
        rhs_type = state.synthesized_types[node.rhs]

        intermediate_type = self.pick_intermediate_type([lhs_type, rhs_type], node.op)
        if intermediate_type is None:
            state.err(
                f"Op {node.op} undefined for {typename(lhs_type)}, {typename(rhs_type)}",
                node,
            )
            return

        if not self.coerce_expr_type(node.lhs, intermediate_type, state):
            return
        if not self.coerce_expr_type(node.rhs, intermediate_type, state):
            return

        result_type = None
        if node.op in NUMERIC_OPERATORS:
            result_type = intermediate_type
        else:
            result_type = BoolValue

        state.op_intermediate_types[node] = intermediate_type
        state.synthesized_types[node] = result_type
        state.contextual_types[node] = result_type

    def visit_AstUnaryOp(self, node: AstUnaryOp, state: CompileState):
        val_type = state.synthesized_types[node.val]

        intermediate_type = self.pick_intermediate_type([val_type], node.op)
        if intermediate_type is None:
            state.err(f"Op {node.op} undefined for {typename(val_type)}", node)
            return

        if not self.coerce_expr_type(node.val, intermediate_type, state):
            return

        result_type = None
        if node.op in NUMERIC_OPERATORS:
            result_type = intermediate_type
        else:
            result_type = BoolValue

        state.op_intermediate_types[node] = intermediate_type
        state.synthesized_types[node] = result_type
        state.contextual_types[node] = result_type

    def visit_AstString(self, node: AstString, state: CompileState):
        state.synthesized_types[node] = FpyStringValue
        state.contextual_types[node] = FpyStringValue

    def visit_AstBoolean(self, node: AstBoolean, state: CompileState):
        state.synthesized_types[node] = BoolValue
        state.contextual_types[node] = BoolValue

    def build_resolved_call_args(
        self,
        node: AstFuncCall,
        func: CallableSymbol,
        node_args: list,
    ) -> list[AstExpr] | CompileError:
        """Build a complete list of argument expressions for a function call.

        This function:
        1. Reorders named arguments to positional order
        2. Fills in default values for missing optional arguments
        3. Checks for missing required arguments

        Returns a list of argument expressions in positional order.
        Returns a CompileError if there's an issue with the arguments.
        """
        func_args = func.args

        # Build a map of parameter name to index
        param_name_to_idx = {arg[0]: i for i, arg in enumerate(func_args)}

        # Track which arguments have been assigned
        assigned_args: list[AstExpr | None] = [None] * len(func_args)
        seen_named = False
        positional_count = 0

        for arg in node_args:
            if is_instance_compat(arg, AstNamedArgument):
                seen_named = True
                # Check if the name is valid
                if arg.name not in param_name_to_idx:
                    return CompileError(
                        f"Unknown argument name '{arg.name}'",
                        arg,
                    )
                idx = param_name_to_idx[arg.name]
                # Check if the argument was already assigned
                if assigned_args[idx] is not None:
                    return CompileError(
                        f"Argument '{arg.name}' specified multiple times",
                        arg,
                    )
                assigned_args[idx] = arg.value
            else:
                # Positional argument
                if seen_named:
                    return CompileError(
                        "Positional argument cannot follow named argument",
                        arg,
                    )
                if positional_count >= len(func_args):
                    return CompileError(
                        f"Too many arguments (expected at most {len(func_args)})",
                        node,
                    )
                # Check if already assigned (shouldn't happen for positional-only case)
                if assigned_args[positional_count] is not None:
                    # This would happen if named arg came before positional
                    return CompileError(
                        f"Argument '{func_args[positional_count][0]}' specified multiple times",
                        arg,
                    )
                assigned_args[positional_count] = arg
                positional_count += 1

        # Fill in default values for missing arguments, error on missing required args
        for i, arg_expr in enumerate(assigned_args):
            if arg_expr is None:
                default_value = func_args[i][2]
                if default_value is not None:
                    assigned_args[i] = default_value
                else:
                    return CompileError(
                        f"Missing required argument '{func_args[i][0]}'",
                        node,
                    )

        return assigned_args

    def check_arg_types_compatible_with_func(
        self,
        node: AstFuncCall,
        func: CallableSymbol,
        resolved_args: list[AstExpr],
        state: CompileState,
    ) -> CompileError | None:
        """Check if a function call's arguments have compatible types.

        Given args must be coercible to expected args, with a special case for casting
        where any numeric type is accepted.
        resolved_args must be in positional order with all values present (defaults filled in).
        Returns a compile error if types don't match, otherwise None.
        """
        func_args = func.args

        if is_instance_compat(func, CastSymbol):
            # casts do not follow coercion rules, because casting is the counterpart of coercion!
            # coercion is implicit, casting is explicit. if they say they want to cast, we let them
            node_arg = resolved_args[0]
            input_type = state.synthesized_types[node_arg]
            output_type = func.to_type
            # right now we only have casting to numbers
            assert output_type in SPECIFIC_NUMERIC_TYPES
            if not issubclass(input_type, NumericalValue):
                # cannot convert a non-numeric type to a numeric type
                return CompileError(
                    f"Expected a number, found {typename(input_type)}", node_arg
                )
            # no error! looks good to me
            return

        # Check provided args against expected
        for value_expr, arg in zip(resolved_args, func_args):
            arg_type = arg[1]

            # Skip type check for default values that are FppValue instances
            # this can only happen if the value is hardcoded into Fpy from a builtin func
            if not is_instance_compat(value_expr, Ast):
                assert is_instance_compat(func, BuiltinSymbol), func
                continue

            # Skip type check for default values from forward-called functions.
            # These expressions haven't been visited yet, so they're not in
            # synthesized_types. Their type compatibility is verified when
            # the function definition is visited.
            if value_expr not in state.synthesized_types:
                continue

            unconverted_type = state.synthesized_types[value_expr]
            if not self.can_coerce_type(unconverted_type, arg_type):
                return CompileError(
                    f"Expected {typename(arg_type)}, found {typename(unconverted_type)}",
                    value_expr if is_instance_compat(value_expr, Ast) else node,
                )
        # all args r good
        return

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        func = state.resolved_symbols.get(node.func)
        if func is None:
            # if it were a reference to a callable, it would have already been resolved
            # if it were a symbol to something else, it would have already errored
            # so it's not even a symbol, just some expr
            state.err(f"Unknown function", node.func)
            return
        node_args = node.args if node.args else []

        # Build resolved args: reorder named args, fill in defaults, check for missing required
        resolved_args = self.build_resolved_call_args(node, func, node_args)
        if is_instance_compat(resolved_args, CompileError):
            state.errors.append(resolved_args)
            return

        # Store the resolved args for use in desugaring and codegen
        state.resolved_func_args[node] = resolved_args

        error_or_none = self.check_arg_types_compatible_with_func(
            node, func, resolved_args, state
        )
        if is_instance_compat(error_or_none, CompileError):
            state.errors.append(error_or_none)
            return
        # otherwise, no error, we're good!

        # okay, we've made sure that the func is possible
        # to call with these args

        # go handle coercion/casting
        if is_instance_compat(func, CastSymbol):
            node_arg = resolved_args[0]
            output_type = func.to_type
            # we're going from input_type to output type, and we're going to ignore
            # the coercion rules
            state.contextual_types[node_arg] = output_type
            # keep track of which ones we explicitly cast. this will
            # let us turn off some checks for boundaries later when we do const folding
            # we turn off the checks because the user is asking us to force this!
            state.expr_explicit_casts.append(node_arg)
        else:
            for value_expr, arg in zip(resolved_args, func.args):
                # Skip coercion for FppValue defaults from builtins
                if not is_instance_compat(value_expr, Ast):
                    assert is_instance_compat(func, BuiltinSymbol), func
                    continue
                # Skip coercion for default values from forward-called functions.
                # These will be coerced when the function definition is visited.
                if value_expr not in state.synthesized_types:
                    continue
                arg_type = arg[1]
                # should be good 2 go based on the check func above
                state.contextual_types[value_expr] = arg_type

        state.synthesized_types[node] = func.return_type
        state.contextual_types[node] = func.return_type

    def visit_AstRange(self, node: AstRange, state: CompileState):
        if not self.coerce_expr_type(node.lower_bound, LoopVarType, state):
            return
        if not self.coerce_expr_type(node.upper_bound, LoopVarType, state):
            return

        state.synthesized_types[node] = RangeValue
        state.contextual_types[node] = RangeValue

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        # should be present in resolved refs because we only let it through if
        # variable is attr, item or var
        lhs_sym = state.resolved_symbols[node.lhs]
        if not is_instance_compat(lhs_sym, (VariableSymbol, FieldSymbol)):
            # assigning to a scope or something
            state.err("Invalid assignment", node.lhs)
            return

        lhs_type = None
        if is_instance_compat(lhs_sym, VariableSymbol):
            lhs_type = lhs_sym.type
        else:
            # reference to a field. make sure that the field is a field of
            # a variable and not like a field of some tlm chan (we can't modify tlm)
            if not is_instance_compat(lhs_sym.base_sym, VariableSymbol):
                state.err("Can only assign variables", node.lhs)
                return
            assert (
                state.contextual_types[node.lhs]
                == state.synthesized_types[node.lhs]
            )
            lhs_type = state.contextual_types[node.lhs]

        # coerce the rhs into the lhs type
        if not self.coerce_expr_type(node.rhs, lhs_type, state):
            return

    def visit_AstAssert(self, node: AstAssert, state: CompileState):
        if not self.coerce_expr_type(node.condition, BoolValue, state):
            return
        if node.exit_code is not None:
            if not self.coerce_expr_type(node.exit_code, U8Value, state):
                return

    def visit_AstFor(self, node: AstFor, state: CompileState):
        # range must coerce to a range!
        if not self.coerce_expr_type(node.range, RangeValue, state):
            return

    def visit_AstWhile(self, node: AstWhile, state: CompileState):
        if not self.coerce_expr_type(node.condition, BoolValue, state):
            return

    def visit_AstIf_AstElif(self, node: Union[AstIf, AstElif], state: CompileState):
        if not self.coerce_expr_type(node.condition, BoolValue, state):
            return

    def visit_AstDef(self, node: AstDef, state: CompileState):
        # Validate that default argument types are compatible with parameter types
        if node.parameters is None:
            return

        func = state.resolved_symbols[node.name]
        if not is_instance_compat(func, FunctionSymbol):
            return

        for (arg_name_var, arg_type_expr, default_value), (_, arg_type, _) in zip(
            node.parameters, func.args
        ):
            if default_value is not None:
                # Check that default value's type can be coerced to parameter type
                if not self.coerce_expr_type(default_value, arg_type, state):
                    return

    def visit_AstReturn(self, node: AstReturn, state: CompileState):
        func = state.enclosing_funcs[node]
        func = state.resolved_symbols[func.name]
        if func.return_type is NothingValue and node.value is not None:
            state.err("Expected no return value", node.value)
            return
        if func.return_type is not NothingValue and node.value is None:
            state.err(
                f"Expected a return value of type {typename(func.return_type)}",
                node.value,
            )
            return
        if node.value is not None:
            if not self.coerce_expr_type(node.value, func.return_type, state):
                return

    def visit_default(self, node, state):
        # coding error, missed an expr
        assert not is_instance_compat(node, AstStmtWithExpr), node


class CalculateDefaultArgConstValues(Visitor):
    """Pass that calculates const values for default argument expressions.

    This must run before CalculateConstExprValues because function call sites may
    reference functions defined later in the source. When we visit a call site that
    uses default arguments, we need the default value's const value to be available.

    This pass also enforces that default values are const expressions.
    """

    def visit_AstDef(self, node: AstDef, state: CompileState):
        if node.parameters is None:
            return

        for arg_name_var, _, default_value in node.parameters:
            if default_value is None:
                continue

            # Run the full CalculateConstExprValues pass on just this default expr
            CalculateConstExprValues().run(default_value, state)
            if len(state.errors) != 0:
                return

            # Check that the default value is a const expression
            const_value = state.contextual_values.get(default_value)
            if const_value is None:
                state.err(
                    f"Default value for argument '{arg_name_var.var}' must be a constant expression",
                    default_value,
                )
                return


class CalculateConstExprValues(Visitor):
    """for each expr, try to calculate its constant value and store it in a map. stores None if no value could be
    calculated at compile time, and NothingType if the expr had no value"""

    @staticmethod
    def _round_float_to_type(value: float, to_type: type[FloatValue]) -> float | None:
        fmt = to_type.get_serialize_format()
        assert fmt is not None, to_type
        try:
            packed = struct.pack(fmt, value)
        except OverflowError:
            return None

        return struct.unpack(fmt, packed)[0]

    @staticmethod
    def _parse_time_string(
        time_str: str, time_base: int, time_context: int, node: Ast, state: CompileState
    ) -> TimeValue | None:
        """Parse an ISO 8601 timestamp string into a TimeValue.

        Accepts formats like:
        - "2025-12-19T14:30:00Z"
        - "2025-12-19T14:30:00.123456Z"

        Returns TimeValue with the provided time_base and time_context, and the parsed
        seconds/microseconds since Unix epoch.
        """
        try:
            # Try parsing with microseconds first
            try:
                dt = datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%S.%fZ")
            except ValueError:
                # Fall back to no microseconds
                dt = datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%SZ")

            # Convert to UTC timestamp
            dt = dt.replace(tzinfo=timezone.utc)
            timestamp = dt.timestamp()

            # Split into seconds and microseconds
            seconds = int(timestamp)
            useconds = int((timestamp - seconds) * 1_000_000)

            # Validate ranges for U32
            if seconds < 0:
                state.err(
                    f"Time string '{time_str}' results in negative seconds ({seconds}), "
                    "which cannot be represented in Fw.Time",
                    node,
                )
                return None
            if seconds > 0xFFFFFFFF:
                state.err(
                    f"Time string '{time_str}' results in seconds ({seconds}) exceeding U32 max",
                    node,
                )
                return None

            return TimeValue(time_base=time_base, time_context=time_context, seconds=seconds, useconds=useconds)

        except ValueError as e:
            state.err(
                f"Invalid time string '{time_str}': expected ISO 8601 format "
                "(e.g., '2025-12-19T14:30:00Z' or '2025-12-19T14:30:00.123456Z')",
                node,
            )
            return None

    @staticmethod
    def const_convert_type(
        from_val: FppValue,
        to_type: FppType,
        node: Ast,
        state: CompileState,
        skip_range_check: bool = False,
    ) -> FppValue | None:
        try:
            from_type = type(from_val)

            if from_type == to_type:
                # no conversion necessary
                return from_val

            if issubclass(to_type, StringValue):
                assert from_type == FpyStringValue, from_type
                return to_type(from_val.val)

            if issubclass(to_type, FloatValue):
                assert issubclass(from_type, NumericalValue), from_type
                from_val = from_val.val

                if to_type == FpyFloatValue:
                    # arbitrary precision
                    # decimal constructor should handle all cases: int, float, or other Decimal
                    return FpyFloatValue(Decimal(from_val))

                # otherwise, we're going to a finite bitwidth float type
                try:
                    coerced_value = float(from_val)
                except OverflowError:
                    state.err(
                        f"{from_val} is out of range for type {typename(to_type)}",
                        node,
                    )
                    return None

                rounded_value = CalculateConstExprValues._round_float_to_type(
                    coerced_value, to_type
                )
                if rounded_value is None:
                    state.err(
                        f"{from_val} is out of range for type {typename(to_type)}",
                        node,
                    )
                    return None

                converted = to_type(rounded_value)
                try:
                    # catch if we would crash the struct packing lib
                    converted.serialize()
                except OverflowError:
                    state.err(
                        f"{from_val} is out of range for type {typename(to_type)}",
                        node,
                    )
                    return None
                return converted
            if issubclass(to_type, IntegerValue):
                assert issubclass(from_type, NumericalValue), from_type
                from_val = from_val.val

                if to_type == FpyIntegerValue:
                    # arbitrary precision
                    # int constructor should handle all cases: int, float, or Decimal
                    return FpyIntegerValue(int(from_val))

                # otherwise going to a finite bitwidth integer type

                if not skip_range_check:
                    # does it fit within bounds?
                    # check that the value can fit in the dest type
                    dest_min, dest_max = to_type.range()
                    if from_val < dest_min or from_val > dest_max:
                        state.err(
                            f"{from_val} is out of range for type {typename(to_type)}",
                            node,
                        )
                        return None

                    # just convert it
                    from_val = int(from_val)
                else:
                    # we skipped the range check, but it's still gotta fit. cut it down

                    # handle narrowing, if necessary
                    from_val = int(from_val)
                    # if signed, convert to unsigned (bit representation should be the same)
                    # first cut down to bitwidth. performed in two's complement
                    mask = (1 << to_type.get_bits()) - 1
                    # this also implicitly converts value to an unsigned number
                    from_val &= mask
                    if to_type in SIGNED_INTEGER_TYPES:
                        # now if the target was signed:
                        sign_bit = 1 << (to_type.get_bits() - 1)
                        if from_val & sign_bit:
                            # the sign bit is set, the result should be negative
                            # subtract the max value as this is how two's complement works
                            from_val -= 1 << to_type.get_bits()

                # okay, we either checked that the value fits in the dest, or we've skipped
                # the check and changed the value to fit
                return to_type(from_val)

            assert False, (from_val, from_type, to_type)
        except TypeException as e:
            state.err(f"For type {typename(from_type)}: {e}", node)
            return None

    def visit_AstLiteral(self, node: AstLiteral, state: CompileState):
        unconverted_type = state.synthesized_types[node]

        try:
            expr_value = unconverted_type(node.value)
        except TypeException as e:
            # TODO can this be reached any more? maybe for string types
            state.err(f"For type {typename(unconverted_type)}: {e}", node)
            return

        skip_range_check = node in state.expr_explicit_casts
        converted_type = state.contextual_types[node]
        if converted_type != unconverted_type:
            expr_value = self.const_convert_type(
                expr_value, converted_type, node, state, skip_range_check
            )
            if expr_value is None:
                return

        state.contextual_values[node] = expr_value

    def visit_AstMemberAccess(self, node: AstMemberAccess, state: CompileState):
        unconverted_type = state.synthesized_types[node]
        converted_type = state.contextual_types[node]
        sym = state.resolved_symbols[node]
        expr_value = None
        if is_instance_compat(sym, (type, dict, CallableSymbol)):
            # these types have no value
            state.contextual_values[node] = NothingValue()
            assert unconverted_type == converted_type, (
                unconverted_type,
                converted_type,
            )
            return
        elif is_instance_compat(sym, (ChTemplate, PrmTemplate, VariableSymbol)):
            # has a value but won't try to calc at compile time
            state.contextual_values[node] = None
            return
        elif is_instance_compat(sym, FppValue):
            expr_value = sym
        elif is_instance_compat(sym, FieldSymbol):
            parent_value = state.contextual_values[node.parent]
            if parent_value is None:
                # no compile time constant value for our parent here
                state.contextual_values[node] = None
                return

            # we are accessing an attribute of something with an fprime value at compile time
            # we must be getting a member
            if is_instance_compat(parent_value, StructValue):
                expr_value = parent_value._val[node.attr]
            elif is_instance_compat(parent_value, TimeValue):
                if node.attr == "seconds":
                    expr_value = U32Value(parent_value.seconds)
                elif node.attr == "useconds":
                    expr_value = U32Value(parent_value.useconds)
                elif node.attr == "time_base":
                    expr_value = U16Value(parent_value.timeBase)
                elif node.attr == "time_context":
                    expr_value = U8Value(parent_value.timeContext)
                else:
                    assert False, node.attr
            else:
                assert False, parent_value

        assert expr_value is not None

        assert is_instance_compat(expr_value, unconverted_type), (
            expr_value,
            unconverted_type,
        )

        skip_range_check = node in state.expr_explicit_casts
        if converted_type != unconverted_type:
            expr_value = self.const_convert_type(
                expr_value, converted_type, node, state, skip_range_check
            )
            if expr_value is None:
                return
        state.contextual_values[node] = expr_value

    def visit_AstIndexExpr(self, node: AstIndexExpr, state: CompileState):
        sym = state.resolved_symbols[node]
        # index expression can only be a field symbol
        assert is_instance_compat(sym, FieldSymbol), sym

        parent_value = state.contextual_values[node.parent]

        if parent_value is None:
            # no compile time constant value for our parent here
            state.contextual_values[node] = None
            return

        assert is_instance_compat(parent_value, ArrayValue), parent_value

        idx = state.contextual_values.get(node.item)
        if idx is None:
            # no compile time constant value for our index
            state.contextual_values[node] = None
            return

        assert is_instance_compat(idx, ArrayIndexType)

        expr_value = parent_value._val[idx._val]

        unconverted_type = state.synthesized_types[node]
        assert is_instance_compat(expr_value, unconverted_type), (
            expr_value,
            unconverted_type,
        )

        skip_range_check = node in state.expr_explicit_casts
        converted_type = state.contextual_types[node]
        if converted_type != unconverted_type:
            expr_value = self.const_convert_type(
                expr_value, converted_type, node, state, skip_range_check
            )
            if expr_value is None:
                return
        state.contextual_values[node] = expr_value

    def visit_AstVar(self, node: AstVar, state: CompileState):
        unconverted_type = state.synthesized_types[node]
        converted_type = state.contextual_types[node]
        sym = state.resolved_symbols[node]
        expr_value = None
        if is_instance_compat(sym, (type, dict, CallableSymbol)):
            # these types have no value
            state.contextual_values[node] = NothingValue()
            assert unconverted_type == converted_type, (
                unconverted_type,
                converted_type,
            )
            return
        elif is_instance_compat(sym, (ChTemplate, PrmTemplate, VariableSymbol)):
            # Has a value but we don't try to calculate it at compile time.
            # NOTE: If you ever add const-folding for VariableSymbol here, you must also
            # update CalculateDefaultArgConstValues. That pass runs CalculateConstExprValues
            # on default argument expressions BEFORE this pass runs on variable assignments.
            # So if a default value references a variable, the variable's const value won't
            # be available yet, and the default value will incorrectly be rejected as non-const.
            state.contextual_values[node] = None
            return
        elif is_instance_compat(sym, FppValue):
            expr_value = sym
        elif is_instance_compat(sym, FieldSymbol):
            assert False, sym

        assert expr_value is not None

        assert is_instance_compat(expr_value, unconverted_type), (
            expr_value,
            unconverted_type,
        )

        skip_range_check = node in state.expr_explicit_casts
        if converted_type != unconverted_type:
            expr_value = self.const_convert_type(
                expr_value, converted_type, node, state, skip_range_check
            )
            if expr_value is None:
                return
        state.contextual_values[node] = expr_value

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        func = state.resolved_symbols[node.func]
        assert is_instance_compat(func, CallableSymbol)

        # Use resolved args from semantic analysis (already in positional order,
        # with defaults filled in)
        # This is guaranteed to be set by PickTypesAndResolveAttrsAndItems
        resolved_args = state.resolved_func_args[node]

        # Gather arg values. Since defaults are already filled in, we just need
        # to look up each arg's const value. For FppValue defaults from builtins,
        # use the value directly.
        arg_values = []
        for arg_expr in resolved_args:
            if is_instance_compat(arg_expr, Ast):
                arg_values.append(state.contextual_values.get(arg_expr))
            else:
                # It's a raw FppValue default from a builtin
                arg_values.append(arg_expr)

        unknown_value = any(v is None for v in arg_values)
        if unknown_value:
            # we will have to calculate this at runtime
            state.contextual_values[node] = None
            return

        expr_value = None

        # whether the conversion that will happen is due to an explicit cast
        if is_instance_compat(func, TypeCtorSymbol):
            # actually construct the type
            if issubclass(func.type, StructValue):
                instance = func.type()
                # pass in args as a dict
                # t[0] is the arg name
                arg_dict = {t[0]: v for t, v in zip(func.type.MEMBER_LIST, arg_values)}
                instance._val = arg_dict
                expr_value = instance

            elif issubclass(func.type, ArrayValue):
                instance = func.type()
                instance._val = arg_values
                expr_value = instance

            elif func.type == TimeValue:
                expr_value = TimeValue(*[val.val for val in arg_values])

            else:
                # no other FppTypees have ctors
                assert False, func.return_type
        elif is_instance_compat(func, CastSymbol):
            # should only be one value. it should be of some numeric type
            # our const convert type func will convert it for us
            expr_value = arg_values[0]
        elif is_instance_compat(func, BuiltinSymbol) and func.name == "time":
            # time() builtin parses ISO 8601 timestamps at compile time
            timestamp_str = arg_values[0].val
            time_base = arg_values[1].val
            time_context = arg_values[2].val
            expr_value = self._parse_time_string(
                timestamp_str, time_base, time_context, node, state
            )
            if expr_value is None:
                return
        else:
            # don't try to calculate the value of this function call
            # it's something like a user defined func, cmd or builtin
            state.contextual_values[node] = None
            return

        unconverted_type = state.synthesized_types[node]
        assert is_instance_compat(expr_value, unconverted_type), (
            expr_value,
            unconverted_type,
        )

        skip_range_check = node in state.expr_explicit_casts
        converted_type = state.contextual_types[node]
        if converted_type != unconverted_type:
            expr_value = self.const_convert_type(
                expr_value, converted_type, node, state, skip_range_check
            )
            if expr_value is None:
                return

        state.contextual_values[node] = expr_value

    def visit_AstBinaryOp(self, node: AstBinaryOp, state: CompileState):
        # Check if both left-hand side (lhs) and right-hand side (rhs) are constants
        lhs_value: FppValue = state.contextual_values.get(node.lhs)
        rhs_value: FppValue = state.contextual_values.get(node.rhs)

        if lhs_value is None or rhs_value is None:
            state.contextual_values[node] = None
            return

        # Both sides are constants, evaluate the operation if the operator is supported

        if not is_instance_compat(lhs_value, ValueType) or not is_instance_compat(
            rhs_value, ValueType
        ):
            # if one of them isn't a ValueType, assume it must be TimeValue
            assert type(lhs_value) == type(rhs_value) and is_instance_compat(
                lhs_value, TimeValue
            ), (
                lhs_value,
                rhs_value,
            )
        else:
            # get the actual pythonic value from the fpp type
            lhs_value = lhs_value.val
            rhs_value = rhs_value.val

        folded_value = None
        # Arithmetic operations
        try:
            if node.op == BinaryStackOp.ADD:
                folded_value = lhs_value + rhs_value
            elif node.op == BinaryStackOp.SUBTRACT:
                folded_value = lhs_value - rhs_value
            elif node.op == BinaryStackOp.MULTIPLY:
                folded_value = lhs_value * rhs_value
            elif node.op == BinaryStackOp.DIVIDE:
                folded_value = lhs_value / rhs_value
            elif node.op == BinaryStackOp.EXPONENT:
                folded_value = lhs_value**rhs_value
            elif node.op == BinaryStackOp.FLOOR_DIVIDE:
                folded_value = lhs_value // rhs_value
            elif node.op == BinaryStackOp.MODULUS:
                folded_value = lhs_value % rhs_value
            # Boolean logic operations
            elif node.op == BinaryStackOp.AND:
                folded_value = lhs_value and rhs_value
            elif node.op == BinaryStackOp.OR:
                folded_value = lhs_value or rhs_value
            # Inequalities
            elif node.op == BinaryStackOp.GREATER_THAN:
                folded_value = lhs_value > rhs_value
            elif node.op == BinaryStackOp.GREATER_THAN_OR_EQUAL:
                folded_value = lhs_value >= rhs_value
            elif node.op == BinaryStackOp.LESS_THAN:
                folded_value = lhs_value < rhs_value
            elif node.op == BinaryStackOp.LESS_THAN_OR_EQUAL:
                folded_value = lhs_value <= rhs_value
            # Equality Checking
            elif node.op == BinaryStackOp.EQUAL:
                if not is_instance_compat(lhs_value, Number):
                    # comparing two complex types
                    assert type(lhs_value) == type(rhs_value), (lhs_value, rhs_value)
                    # for now we don't fold this
                    folded_value = None
                else:
                    folded_value = lhs_value == rhs_value
            elif node.op == BinaryStackOp.NOT_EQUAL:
                if not is_instance_compat(lhs_value, Number):
                    # comparing two complex types
                    assert type(lhs_value) == type(rhs_value), (lhs_value, rhs_value)
                    # for now we don't fold this
                    folded_value = None
                else:
                    folded_value = lhs_value != rhs_value
            else:
                # missing an operation
                assert False, node.op
        except ZeroDivisionError:
            state.err("Divide by zero error", node)
            return
        except OverflowError:
            state.err("Overflow error", node)
            return
        except ValueError as err:
            state.err(str(err) if str(err) else "Domain error", node)
            return
        except decimal.InvalidOperation:
            state.err("Domain error", node)
            return

        if folded_value is None:
            # give up, don't try to calculate the value of this expr at compile time
            state.contextual_values[node] = None
            return

        if type(folded_value) == int:
            folded_value = FpyIntegerValue(folded_value)
        elif type(folded_value) == Decimal:
            folded_value = FpyFloatValue(folded_value)
        elif type(folded_value) == bool:
            folded_value = BoolValue(folded_value)
        else:
            assert False, folded_value

        # first fold, store the result in arbitrary precision

        # then if the expression is some other type, convert:
        skip_range_check = node in state.expr_explicit_casts
        unconverted_type = state.synthesized_types.get(node)
        # the intent of this is to handle situations where we're constant folding and the results cannot be arbitrary precision
        folded_value = self.const_convert_type(
            folded_value, unconverted_type, node, state, skip_range_check=False
        )

        converted_type = state.contextual_types.get(node)
        # okay and now perform type coercion/casting
        if converted_type != unconverted_type:
            folded_value = self.const_convert_type(
                folded_value, converted_type, node, state, skip_range_check
            )
            if folded_value is None:
                return
        state.contextual_values[node] = folded_value

    def visit_AstUnaryOp(self, node: AstUnaryOp, state: CompileState):
        value: FppValue = state.contextual_values.get(node.val)

        if value is None:
            state.contextual_values[node] = None
            return

        # input is constant, evaluate the operation if the operator is supported
        assert is_instance_compat(value, ValueType), value

        # get the actual pythonic value from the fpp type
        value = value.val
        folded_value = None

        if node.op == UnaryStackOp.NEGATE:
            folded_value = -value
        elif node.op == UnaryStackOp.IDENTITY:
            folded_value = value
        elif node.op == UnaryStackOp.NOT:
            folded_value = not value
        else:
            # missing an operation
            assert False, node.op

        assert folded_value is not None

        if type(folded_value) == int:
            folded_value = FpyIntegerValue(folded_value)
        elif type(folded_value) == Decimal:
            folded_value = FpyFloatValue(folded_value)
        elif type(folded_value) == bool:
            folded_value = BoolValue(folded_value)
        else:
            assert False, folded_value

        # first fold, store the result in arbitrary precision

        # then if the expression is some other type, convert:
        skip_range_check = node in state.expr_explicit_casts
        unconverted_type = state.synthesized_types.get(node)
        # the intent of this is to handle situations where we're constant folding and the results cannot be arbitrary precision
        folded_value = self.const_convert_type(
            folded_value, unconverted_type, node, state, skip_range_check=False
        )

        converted_type = state.contextual_types.get(node)
        if converted_type != unconverted_type:
            folded_value = self.const_convert_type(
                folded_value, converted_type, node, state, skip_range_check
            )
            if folded_value is None:
                return
        state.contextual_values[node] = folded_value

    def visit_AstRange(self, node: AstRange, state: CompileState):
        # ranges don't really end up having a value, they kinda just exist as a type
        state.contextual_values[node] = None

    def visit_default(self, node, state):
        # coding error, missed an expr
        assert not is_instance_compat(node, AstExpr), node


class CheckAllBranchesReturn(Visitor):
    def visit_AstReturn(self, node: AstReturn, state: CompileState):
        state.does_return[node] = True

    def visit_AstStmtList(self, node: Union[AstStmtList, AstBlock], state: CompileState):
        state.does_return[node] = any(state.does_return[n] for n in node.stmts)

    def visit_AstIf(self, node: AstIf, state: CompileState):
        # an if statement returns if all of its branches return
        branch_returns = [state.does_return[node.body]]

        for _elif in node.elifs:
            branch_returns.append(state.does_return[_elif])

        if node.els is not None:
            branch_returns.append(state.does_return[node.els])
        else:
            # implicit else branch that falls through without returning
            branch_returns.append(False)

        state.does_return[node] = all(branch_returns)

    def visit_AstElif(self, node: Union[AstElif], state: CompileState):
        state.does_return[node] = state.does_return[node.body]

    def visit_AstDef(self, node: AstDef, state: CompileState):
        # if we found another func def inside this body, it definitely doesn't return
        state.does_return[node] = False

    def visit_AstAssign_AstPass_AstAssert_AstContinue_AstBreak_AstWhile_AstFor(
        self,
        node: Union[
            AstAssign, AstPass, AstAssert, AstContinue, AstBreak, AstWhile, AstFor
        ],
        state: CompileState,
    ):
        state.does_return[node] = False

    def visit_AstExpr(self, node: AstExpr, state: CompileState):
        # expressions do not return
        state.does_return[node] = False

    def visit_default(self, node, state):
        assert not is_instance_compat(node, AstStmt)


class CheckFunctionReturns(Visitor):
    def visit_AstDef(self, node: AstDef, state: CompileState):
        CheckAllBranchesReturn().run(node.body, state)
        if node.return_type is None:
            # don't need to return explicitly
            return
        if not state.does_return[node.body]:
            state.err(
                f"Function '{node.name.var}' does not always return a value", node
            )
            return


class CheckConstArrayAccesses(Visitor):
    def visit_AstIndexExpr(self, node: AstIndexExpr, state: CompileState):
        # if the index is a const, we should be able to check if it's in bounds
        idx_value = state.contextual_values.get(node.item)
        if idx_value is None:
            # can't check at compile time
            return

        parent_type = state.contextual_types[node.parent]
        assert issubclass(parent_type, ArrayValue), parent_type

        if idx_value.val < 0 or idx_value.val >= parent_type.LENGTH:
            state.err(
                f"Index {idx_value.val} out of bounds for array type {typename(parent_type)} with length {parent_type.LENGTH}",
                node.item,
            )
            return


class WarnRangesAreNotEmpty(Visitor):
    def visit_AstRange(self, node: AstRange, state: CompileState):
        # if the index is a const, we should be able to check if it's in bounds
        lower_value: LoopVarType = state.contextual_values.get(node.lower_bound)
        upper_value: LoopVarType = state.contextual_values.get(node.upper_bound)
        if lower_value is None or upper_value is None:
            # cannot check at compile time
            return

        if lower_value.val >= upper_value.val:
            state.warn("Range is empty", node)
